"""Unit tests for sonification methods"""

import pytest
import mir_eval
import numpy as np
import scipy


@pytest.mark.parametrize("times", [np.array([1.0]), np.arange(10.0)])
@pytest.mark.parametrize("fs", [8000, 44100])
def test_clicks(times, fs):
    # Test output length for a variety of parameter settings
    click_signal = mir_eval.sonify.clicks(times, fs)
    assert len(click_signal) == times.max() * fs + int(fs * 0.1) + 1
    click_signal = mir_eval.sonify.clicks(times, fs, length=1000)
    assert len(click_signal) == 1000
    click_signal = mir_eval.sonify.clicks(times, fs, click=np.zeros(1000))
    assert len(click_signal) == times.max() * fs + 1000 + 1


@pytest.mark.parametrize("fs", [8000, 44100])
def test_time_frequency(fs):
    # Test length for different inputs
    signal = mir_eval.sonify.time_frequency(
        np.random.standard_normal((100, 1000)),
        np.arange(1, 101),
        np.linspace(0, 10, 1000),
        fs,
    )
    assert len(signal) == 10 * fs

    # Make one longer
    signal = mir_eval.sonify.time_frequency(
        np.random.standard_normal((100, 1000)),
        np.arange(1, 101),
        np.linspace(0, 10, 1000),
        fs,
        length=fs * 11,
    )
    assert len(signal) == 11 * fs

    # Make one shorter
    signal = mir_eval.sonify.time_frequency(
        np.random.standard_normal((100, 1000)),
        np.arange(1, 101),
        np.linspace(0, 10, 1000),
        fs,
        length=fs * 5,
    )
    assert len(signal) == 5 * fs


def test_time_frequency_const():

    # Sonify with a single interval to hit the const interpolator
    s1 = mir_eval.sonify.time_frequency(
        np.ones((1, 1)),
        np.array([60]),
        np.array([[0, 1]]),
        8000,
    )
    # Sonify with two tiem intervals to hit the regular interpolator
    # but the second frequency will have no energy, so it doesn't
    # change the resulting signal
    s2 = mir_eval.sonify.time_frequency(
        np.array([[1, 1], [0, 0]]),
        np.array([60, 90]),
        np.array([[0, 1], [0, 1]]),
        8000,
    )

    assert np.allclose(s1, s2)


def test_time_frequency_offset():
    fs = 8000
    # Length is 3 seconds, first interval starts at 5.
    # Should produce an empty signal.
    signal = mir_eval.sonify.time_frequency(
        np.random.standard_normal((100, 100)),
        np.arange(1, 101),
        np.linspace(5, 10, 100),
        fs,
        length=fs * 3,
    )
    assert len(signal) == 3 * fs
    assert np.allclose(signal, 0)


@pytest.mark.xfail(raises=ValueError)
def test_time_frequency_badtime():
    fs = 8000
    gram = np.ones((10, 20))
    times = np.arange(10)

    mir_eval.sonify.time_frequency(gram, np.arange(1, 11), times, fs)


@pytest.mark.xfail(raises=ValueError)
def test_time_frequency_badintervals():
    fs = 8000
    gram = np.ones((10, 20))
    times = np.ones((11, 2))

    mir_eval.sonify.time_frequency(gram, np.arange(1, 11), times, fs)


@pytest.mark.xfail(raises=ValueError)
def test_time_frequency_frequency():
    fs = 8000
    gram = np.ones((10, 20))
    times = np.arange(20)

    mir_eval.sonify.time_frequency(gram, np.arange(1, 8), times, fs)


@pytest.mark.parametrize("fs", [8000, 44100])
def test_chroma(fs):
    signal = mir_eval.sonify.chroma(
        np.random.standard_normal((12, 1000)), np.linspace(0, 10, 1000), fs
    )
    assert len(signal) == 10 * fs
    signal = mir_eval.sonify.chroma(
        np.random.standard_normal((12, 1000)),
        np.linspace(0, 10, 1000),
        fs,
        length=fs * 11,
    )
    assert len(signal) == 11 * fs


@pytest.mark.parametrize("fs", [8000, 44100])
def test_chords(fs):
    intervals = np.array([np.arange(10), np.arange(1, 11)]).T
    signal = mir_eval.sonify.chords(
        ["C", "C:maj", "D:min7", "E:min", "C#", "C", "C", "C", "C", "C"], intervals, fs
    )
    assert len(signal) == 10 * fs
    signal = mir_eval.sonify.chords(
        ["C", "C:maj", "D:min7", "E:min", "C#", "C", "C", "C", "C", "C"],
        intervals,
        fs,
        length=fs * 11,
    )
    assert len(signal) == 11 * fs


def test_chord_x():
    # This test verifies that X sonifies as silence
    intervals = np.array([[0, 1]])
    signal = mir_eval.sonify.chords(["X"], intervals, 8000)
    assert not np.any(signal), signal


def test_pitch_contour():
    # Generate some random pitch
    fs = 8000
    times = np.linspace(0, 5, num=5 * fs, endpoint=True)

    noise = scipy.ndimage.gaussian_filter1d(np.random.randn(len(times)), sigma=256)
    freqs = 440.0 * 2.0 ** (16 * noise)
    amps = np.linspace(0, 1, num=5 * fs, endpoint=True)

    # negate a bunch of sequences
    idx = np.unique(np.random.randint(0, high=len(times), size=32))
    for start, end in zip(idx[::2], idx[1::2]):
        freqs[start:end] *= -1

    # Test with inferring duration
    x = mir_eval.sonify.pitch_contour(times, freqs, fs)
    assert len(x) == fs * 5

    # Test with an explicit duration
    # This forces the interpolator to go off the end of the sampling grid,
    # which should result in a constant sequence in the output
    x = mir_eval.sonify.pitch_contour(times, freqs, fs, length=fs * 7)
    assert len(x) == fs * 7
    assert np.allclose(x[-fs * 2 :], x[-fs * 2])

    # Test with an explicit duration and a fixed offset
    # This forces the interpolator to go off the beginning of
    # the sampling grid, which should result in a constant output
    x = mir_eval.sonify.pitch_contour(times + 5.0, freqs, fs, length=fs * 7)
    assert len(x) == fs * 7
    assert np.allclose(x[: fs * 5], x[0])

    # Test with explicit amplitude
    x = mir_eval.sonify.pitch_contour(times, freqs, fs, length=fs * 7, amplitudes=amps)
    assert len(x) == fs * 7
    assert np.allclose(x[0], 0)



"""Unit tests for input/output functions"""

import numpy as np
import json
import mir_eval
import warnings
import pathlib
import tempfile
import pytest


def test_load_delimited():
    # Test for ValueError when a non-string or file handle is passed
    with pytest.raises(IOError):
        mir_eval.io.load_delimited(None, [int])

    # Test for a value error when the wrong number of columns is passed
    with tempfile.TemporaryFile("r+") as f:
        f.write("10 20")
        f.seek(0)
        with pytest.raises(ValueError):
            mir_eval.io.load_delimited(f, [int, int, int])

    # Test for a value error on conversion failure
    with tempfile.TemporaryFile("r+") as f:
        f.write("10 a 30")
        f.seek(0)
        with pytest.raises(ValueError):
            mir_eval.io.load_delimited(f, [int, int, int])


def test_load_delimited_commented():
    with tempfile.TemporaryFile("r+") as f:
        f.write("; some comment\n10 20\n30 50")
        f.seek(0)
        col1, col2 = mir_eval.io.load_delimited(f, [int, int], comment=";")
        assert np.allclose(col1, [10, 30])
        assert np.allclose(col2, [20, 50])

        # Rewind and try with the default comment character
        f.seek(0)
        with pytest.raises(ValueError):
            mir_eval.io.load_delimited(f, [int, int])

        # Rewind and try with no comment support
        f.seek(0)
        with pytest.raises(ValueError):
            mir_eval.io.load_delimited(f, [int, int], comment=None)


def test_load_delimited_nocomment():
    with tempfile.TemporaryFile("r+") as f:
        f.write("10 20\n30 50")
        f.seek(0)
        col1, col2 = mir_eval.io.load_delimited(f, [int, int])
        assert np.allclose(col1, [10, 30])
        assert np.allclose(col2, [20, 50])

        # Rewind and try with a different comment char
        f.seek(0)
        col1, col2 = mir_eval.io.load_delimited(f, [int, int], comment=";")
        assert np.allclose(col1, [10, 30])
        assert np.allclose(col2, [20, 50])

        # Rewind and try with no different comment string
        f.seek(0)
        col1, col2 = mir_eval.io.load_delimited(f, [int, int], comment=None)
        assert np.allclose(col1, [10, 30])
        assert np.allclose(col2, [20, 50])


@pytest.mark.parametrize(
    "file_or_path", ["data/beat/ref01.txt", pathlib.Path("data/beat/ref01.txt")]
)
def test_load_delimited_file_or_path(file_or_path):
    # Tests the load_delimited routine with either a string or a
    # pathlib.Path object as the first argument
    data = mir_eval.io.load_delimited(file_or_path, [float])
    assert len(data) == 635


def test_load_events():
    # Test for a warning when invalid events are supplied
    with tempfile.TemporaryFile("r+") as f:
        with pytest.warns(UserWarning, match="Events should be in increasing order."):
            # Non-increasing is invalid
            f.write("10\n9")
            f.seek(0)
            events = mir_eval.io.load_events(f)
            # Make sure events were read in correctly
            assert np.all(events == [10, 9])


def test_load_labeled_events():
    # Test for a value error when invalid labeled events are supplied
    with tempfile.TemporaryFile("r+") as f:
        with pytest.warns(UserWarning, match="Events should be in increasing order."):
            # Non-increasing is invalid
            f.write("10 a\n9 b")
            f.seek(0)
            events, labels = mir_eval.io.load_labeled_events(f)
            # Make sure events were read in correctly
            assert np.all(events == [10, 9])
            # Make sure labels were read in correctly
            assert labels == ["a", "b"]


def test_load_intervals():
    # Test for a value error when invalid labeled events are supplied
    with tempfile.TemporaryFile("r+") as f:
        with pytest.warns(
            UserWarning, match="All interval durations must be strictly positive"
        ):
            # Non-increasing is invalid
            f.write("10 9\n9 10")
            f.seek(0)
            intervals = mir_eval.io.load_intervals(f)
            # Make sure intervals were read in correctly
            assert np.all(intervals == [[10, 9], [9, 10]])


def test_load_labeled_intervals():
    # Test for a value error when invalid labeled events are supplied
    with tempfile.TemporaryFile("r+") as f:
        with pytest.warns(
            UserWarning, match="All interval durations must be strictly positive"
        ):
            # Non-increasing is invalid
            f.write("10 9 a\n9 10 b")
            f.seek(0)
            intervals, labels = mir_eval.io.load_labeled_intervals(f)
            # Make sure intervals were read in correctly
            assert np.all(intervals == [[10, 9], [9, 10]])
            assert labels == ["a", "b"]


def test_load_valued_intervals():
    # Test for a value error when invalid valued events are supplied
    with tempfile.TemporaryFile("r+") as f:
        with pytest.warns(
            UserWarning, match="All interval durations must be strictly positive"
        ):
            # Non-increasing is invalid
            f.write("10 9 5\n9 10 6")
            f.seek(0)
            intervals, values = mir_eval.io.load_valued_intervals(f)
            # Make sure intervals were read in correctly
            assert np.all(intervals == [[10, 9], [9, 10]])
            assert np.all(values == [5, 6])


def test_load_ragged_time_series():
    # Test for ValueError when a non-string or file handle is passed
    with pytest.raises(IOError):
        mir_eval.io.load_ragged_time_series(None, float, header=False)
    # Test for a value error on conversion failure
    with tempfile.TemporaryFile("r+") as f:
        f.write("10 a 30")
        f.seek(0)
        with pytest.raises(ValueError):
            mir_eval.io.load_ragged_time_series(f, float, header=False)
    # Test for a value error on invalid time stamp
    with tempfile.TemporaryFile("r+") as f:
        f.write("a 10 30")
        f.seek(0)
        with pytest.raises(ValueError):
            mir_eval.io.load_ragged_time_series(f, int, header=False)
    # Test for a value error on invalid time stamp with header
    with tempfile.TemporaryFile("r+") as f:
        f.write("x y z\na 10 30")
        f.seek(0)
        with pytest.raises(ValueError):
            mir_eval.io.load_ragged_time_series(f, int, header=True)

    with tempfile.TemporaryFile("r+") as f:
        f.write("#comment\n0 1 2\n3 4\n# comment\n5 6 7")
        f.seek(0)
        times, values = mir_eval.io.load_ragged_time_series(
            f, int, header=False, comment="#"
        )
        assert np.allclose(times, [0, 3, 5])
        assert np.allclose(values[0], [1, 2])
        assert np.allclose(values[1], [4])
        assert np.allclose(values[2], [6, 7])

        # Rewind with a wrong comment string
        f.seek(0)
        with pytest.raises(ValueError):
            mir_eval.io.load_ragged_time_series(f, int, header=False, comment="%")

        # Rewind with no comment string
        f.seek(0)
        with pytest.raises(ValueError):
            mir_eval.io.load_ragged_time_series(f, int, header=False, comment=None)


def test_load_tempo():
    # Test the tempo loader
    tempi, weight = mir_eval.io.load_tempo("data/tempo/ref01.lab")

    assert np.allclose(tempi, [60, 120])
    assert weight == 0.5


@pytest.mark.xfail(raises=ValueError)
def test_load_tempo_multiline():
    tempi, weight = mir_eval.io.load_tempo("data/tempo/bad00.lab")


@pytest.mark.xfail(raises=ValueError)
def test_load_tempo_badweight():
    tempi, weight = mir_eval.io.load_tempo("data/tempo/bad01.lab")


def test_load_bad_tempi():
    with pytest.warns(UserWarning, match="non-negative numbers"):
        tempi, weight = mir_eval.io.load_tempo("data/tempo/bad02.lab")



import pytest
import mir_eval
import numpy as np
import glob
import json

A_TOL = 1e-12

# Path to the fixture files
REF_GLOB = "data/transcription_velocity/ref*.txt"
EST_GLOB = "data/transcription_velocity/est*.txt"
SCORES_GLOB = "data/transcription_velocity/output*.json"
ref_files = sorted(glob.glob(REF_GLOB))
est_files = sorted(glob.glob(EST_GLOB))
sco_files = sorted(glob.glob(SCORES_GLOB))

assert len(ref_files) == len(est_files) == len(sco_files) > 0

file_sets = list(zip(ref_files, est_files, sco_files))


def _load_transcription_velocity(filename):
    """Loader for data in the format start, end, pitch, velocity."""
    starts, ends, pitches, velocities = mir_eval.io.load_delimited(
        filename, [float, float, int, int]
    )
    # Stack into an interval matrix
    intervals = np.array([starts, ends]).T
    # return pitches and velocities as np.ndarray
    pitches = np.array(pitches)
    velocities = np.array(velocities)
    return intervals, pitches, velocities


@pytest.fixture
def velocity_data(request):
    ref_f, est_f, sco_f = request.param
    with open(sco_f) as f:
        expected_scores = json.load(f)
    # Load in reference transcription
    ref_int, ref_pitch, ref_vel = _load_transcription_velocity(ref_f)
    # Load in estimated transcription
    est_int, est_pitch, est_vel = _load_transcription_velocity(est_f)

    return (ref_int, ref_pitch, ref_vel), (est_int, est_pitch, est_vel), expected_scores


def test_negative_velocity():
    good_i, good_p, good_v = np.array([[0, 1]]), np.array([100]), np.array([1])
    bad_i, bad_p, bad_v = np.array([[0, 1]]), np.array([100]), np.array([-1])

    with pytest.raises(ValueError):
        mir_eval.transcription_velocity.validate(
            bad_i, bad_p, bad_v, good_i, good_p, good_v
        )
    with pytest.raises(ValueError):
        mir_eval.transcription_velocity.validate(
            good_i, good_p, good_v, bad_i, bad_p, bad_v
        )


def test_wrong_shape_velocity():
    good_i, good_p, good_v = np.array([[0, 1]]), np.array([100]), np.array([1])
    bad_i, bad_p, bad_v = np.array([[0, 1]]), np.array([100]), np.array([1, 2])

    with pytest.raises(ValueError):
        mir_eval.transcription_velocity.validate(
            bad_i, bad_p, bad_v, good_i, good_p, good_v
        )
    with pytest.raises(ValueError):
        mir_eval.transcription_velocity.validate(
            good_i, good_p, good_v, bad_i, bad_p, bad_v
        )


def test_precision_recall_f1_overlap():
    # Simple unit test
    ref_i = np.array([[0, 1], [0.5, 0.7], [1, 2]])
    ref_p = np.array([100, 110, 80])
    ref_v = np.array([10, 90, 110])
    est_i = np.array([[0, 1], [0.5, 0.7], [1, 2]])
    est_p = np.array([100, 110, 80])
    est_v = np.array([10, 70, 110])
    p, r, f, o = mir_eval.transcription_velocity.precision_recall_f1_overlap(
        ref_i, ref_p, ref_v, est_i, est_p, est_v
    )
    assert np.allclose((p, r, f, o), (2 / 3.0, 2 / 3.0, 2 / 3.0, 1.0))
    p, r, f, o = mir_eval.transcription_velocity.precision_recall_f1_overlap(
        ref_i, ref_p, ref_v, est_i, est_p, est_v, velocity_tolerance=0.3
    )
    assert np.allclose((p, r, f, o), (1.0, 1.0, 1.0, 1.0))


# Suppressing this warning. We know the notes are empty, that's not the point.
@pytest.mark.filterwarnings("ignore:.*notes are empty")
def test_precision_recall_f1_overlap_empty():
    good_i, good_p, good_v = np.array([[0, 1]]), np.array([100]), np.array([1])
    bad_i, bad_p, bad_v = np.empty((0, 2)), np.array([]), np.array([])
    p, r, f, o = mir_eval.transcription_velocity.precision_recall_f1_overlap(
        good_i, good_p, good_v, bad_i, bad_p, bad_v
    )
    assert (p, r, f, o) == (0.0, 0.0, 0.0, 0.0)
    p, r, f, o = mir_eval.transcription_velocity.precision_recall_f1_overlap(
        bad_i, bad_p, bad_v, good_i, good_p, good_v
    )
    assert (p, r, f, o) == (0.0, 0.0, 0.0, 0.0)


def test_precision_recall_f1_overlap_no_overlap():
    p, r, f, o = mir_eval.transcription_velocity.precision_recall_f1_overlap(
        np.array([[1, 2]]),
        np.array([1]),
        np.array([1]),
        np.array([[3, 4]]),
        np.array([1]),
        np.array([1]),
    )
    assert (p, r, f, o) == (0.0, 0.0, 0.0, 0.0)


@pytest.mark.parametrize("velocity_data", file_sets, indirect=True)
def test_regression(velocity_data):
    (
        (ref_int, ref_pitch, ref_vel),
        (est_int, est_pitch, est_vel),
        expected_scores,
    ) = velocity_data

    scores = mir_eval.transcription_velocity.evaluate(
        ref_int, ref_pitch, ref_vel, est_int, est_pitch, est_vel
    )
    assert scores.keys() == expected_scores.keys()
    for metric in scores:
        assert np.allclose(scores[metric], expected_scores[metric], atol=A_TOL), metric



"""
Unit tests for mir_eval.beat
"""

import numpy as np
import json
import mir_eval
import glob
import pytest

A_TOL = 1e-12

# Path to the fixture files
REF_GLOB = "data/beat/ref*.txt"
EST_GLOB = "data/beat/est*.txt"
SCORES_GLOB = "data/beat/output*.json"

ref_files = sorted(glob.glob(REF_GLOB))
est_files = sorted(glob.glob(EST_GLOB))
sco_files = sorted(glob.glob(SCORES_GLOB))

assert len(ref_files) == len(est_files) == len(sco_files) > 0

file_sets = list(zip(ref_files, est_files, sco_files))


@pytest.fixture
def beat_data(request):
    ref_f, est_f, sco_f = request.param
    with open(sco_f) as f:
        expected_scores = json.load(f)
    reference_beats = mir_eval.io.load_events(ref_f)
    estimated_beats = mir_eval.io.load_events(est_f)

    return reference_beats, estimated_beats, expected_scores


def test_trim_beats():
    # Construct dummy beat times [0., 1., ...]
    dummy_beats = np.arange(10, dtype=np.float64)
    # We expect trim_beats to remove all beats < 5s
    expected_beats = np.arange(5, 10, dtype=np.float64)
    assert np.allclose(mir_eval.beat.trim_beats(dummy_beats), expected_beats)


@pytest.mark.parametrize(
    "metric",
    [
        mir_eval.beat.f_measure,
        mir_eval.beat.cemgil,
        mir_eval.beat.goto,
        mir_eval.beat.p_score,
        mir_eval.beat.continuity,
        mir_eval.beat.information_gain,
    ],
)
def test_beat_empty_warnings(metric):
    with pytest.warns(UserWarning, match="Reference beats are empty."):
        metric(np.array([]), np.arange(10))
    with pytest.warns(UserWarning, match="Estimated beats are empty."):
        metric(np.arange(10), np.array([]))
    with pytest.warns(UserWarning, match="beats are empty."):
        assert np.allclose(metric(np.array([]), np.array([])), 0)


@pytest.mark.parametrize(
    "metric",
    [
        mir_eval.beat.f_measure,
        mir_eval.beat.cemgil,
        mir_eval.beat.goto,
        mir_eval.beat.p_score,
        mir_eval.beat.continuity,
        mir_eval.beat.information_gain,
    ],
)
@pytest.mark.parametrize(
    "beats",
    [
        np.array([[1.0, 2.0]]),  # beats must be a 1d array
        np.array([1e10, 1e11]),  # beats must be not huge
        np.array([2.0, 1.0]),  # must be sorted
    ],
)
@pytest.mark.xfail(raises=ValueError)
def test_beat_fail(metric, beats):
    metric(beats, beats)


@pytest.mark.parametrize(
    "metric",
    [
        mir_eval.beat.f_measure,
        mir_eval.beat.cemgil,
        mir_eval.beat.goto,
        mir_eval.beat.p_score,
        mir_eval.beat.continuity,
        mir_eval.beat.information_gain,
    ],
)
def test_beat_perfect(metric):
    beats = np.arange(10, dtype=np.float64)
    assert np.allclose(metric(beats, beats), 1)


@pytest.mark.parametrize("beat_data", file_sets, indirect=True)
def test_beat_functions(beat_data):
    reference_beats, estimated_beats, expected_scores = beat_data

    # Compute scores
    scores = mir_eval.beat.evaluate(reference_beats, estimated_beats)
    # Compare them
    assert scores.keys() == expected_scores.keys()
    for metric in scores:
        assert np.allclose(scores[metric], expected_scores[metric], atol=A_TOL)


# Unit tests for specific behavior not covered by the above
def test_goto_proportion_correct():
    # This covers the case when over 75% of the beat tracking is correct, and
    # more than 3 beats are incorrect
    assert mir_eval.beat.goto(
        np.arange(100), np.append(np.arange(80), np.arange(80, 100) + 0.2)
    )


@pytest.mark.parametrize(
    "metric",
    [mir_eval.beat.p_score, mir_eval.beat.continuity, mir_eval.beat.information_gain],
)
def test_warning_on_one_beat(metric):
    # This tests the metrics where passing only a single beat raises a warning
    # and returns 0

    with pytest.warns(UserWarning, match="Only one reference beat"):
        metric(np.array([10]), np.arange(10))
    with pytest.warns(UserWarning, match="Only one estimated beat"):
        metric(np.arange(10), np.array([10]))


def test_continuity_edge_cases():
    # There is some special-case logic for when there are few beats
    assert np.allclose(
        mir_eval.beat.continuity(np.array([6.0, 6.0]), np.array([6.0, 7.0])), 0.0
    )
    assert np.allclose(
        mir_eval.beat.continuity(np.array([6.0, 6.0]), np.array([6.5, 7.0])), 0.0
    )



"""
Unit tests for mir_eval.segment
"""

import numpy as np
import json
import mir_eval
import glob
import pytest

A_TOL = 1e-12

# Path to the fixture files
REF_GLOB = "data/segment/ref*.lab"
EST_GLOB = "data/segment/est*.lab"
SCORES_GLOB = "data/segment/output*.json"

ref_files = sorted(glob.glob(REF_GLOB))
est_files = sorted(glob.glob(EST_GLOB))
sco_files = sorted(glob.glob(SCORES_GLOB))

assert len(ref_files) == len(est_files) == len(sco_files) > 0

file_sets = list(zip(ref_files, est_files, sco_files))


@pytest.fixture
def segment_data(request):
    ref_f, est_f, sco_f = request.param
    with open(sco_f) as f:
        expected_scores = json.load(f)
    # Load in an example segmentation annotation
    ref_intervals, ref_labels = mir_eval.io.load_labeled_intervals(ref_f)
    # Load in an example segmentation tracker output
    est_intervals, est_labels = mir_eval.io.load_labeled_intervals(est_f)

    return ref_intervals, ref_labels, est_intervals, est_labels, expected_scores


@pytest.mark.parametrize(
    "metric", [mir_eval.segment.detection, mir_eval.segment.deviation]
)
def test_segment_boundary_empty(metric):
    with pytest.warns(UserWarning, match="Reference intervals are empty"):
        metric(np.zeros((0, 2)), np.array([[1, 2], [2, 3]]), trim=False)

    with pytest.warns(UserWarning, match="Estimated intervals are empty"):
        metric(np.array([[1, 2], [2, 3]]), np.array([[1, 2]]), trim=True)

    with pytest.warns(UserWarning, match="intervals are empty"):
        empty_intervals = np.zeros((0, 2))
        if metric == mir_eval.segment.detection:
            assert np.allclose(metric(empty_intervals, empty_intervals), 0)
        else:
            assert np.all(np.isnan(metric(empty_intervals, empty_intervals)))


@pytest.mark.xfail(raises=ValueError)
@pytest.mark.parametrize(
    "metric", [mir_eval.segment.detection, mir_eval.segment.deviation]
)
@pytest.mark.parametrize(
    "intervals",
    [
        # Now test validation function - intervals must be n by 2
        np.array([1, 2, 3, 4]),
        # Interval boundaries must be positive
        np.array([[-1, 2], [2, 3]]),
        # Positive interval durations
        np.array([[2, 1], [2, 3]]),
    ],
)
def test_segment_boundary_errors(metric, intervals):
    metric(intervals, intervals)


def test_segment_boundary_detection_perfect():
    correct_intervals = np.array([[0, 1], [1, 2]])
    assert np.allclose(
        mir_eval.segment.detection(correct_intervals, correct_intervals), 1
    )


def test_segment_boundary_deviation_perfect():
    correct_intervals = np.array([[0, 1], [1, 2]])
    assert np.allclose(
        mir_eval.segment.deviation(correct_intervals, correct_intervals), 0
    )


@pytest.mark.parametrize(
    "metric",
    [
        mir_eval.segment.pairwise,
        mir_eval.segment.rand_index,
        mir_eval.segment.ari,
        mir_eval.segment.mutual_information,
        mir_eval.segment.nce,
        mir_eval.segment.vmeasure,
    ],
)
def test_segment_structure_empty(metric):
    with pytest.warns(UserWarning, match="Reference intervals are empty"):
        metric(np.zeros((0, 2)), [], np.array([[0, 1]]), ["foo"])

    with pytest.warns(UserWarning, match="Estimated intervals are empty"):
        metric(np.array([[0, 1]]), ["foo"], np.zeros((0, 2)), [])

    with pytest.warns(UserWarning, match="intervals are empty"):
        empty_intervals = np.zeros((0, 2))
        assert np.allclose(metric(empty_intervals, [], empty_intervals, []), 0)


@pytest.mark.xfail(raises=ValueError)
@pytest.mark.parametrize(
    "metric",
    [
        mir_eval.segment.pairwise,
        mir_eval.segment.rand_index,
        mir_eval.segment.ari,
        mir_eval.segment.mutual_information,
        mir_eval.segment.nce,
        mir_eval.segment.vmeasure,
    ],
)
@pytest.mark.parametrize(
    "intervals, labels",
    [
        # Test for non-matching numbers of intervals and labels
        (np.array([[2, 1], [2, 3]]), ["a", "b", "c"]),
        # Now test validation function - intervals must be n by 2
        (np.arange(4), ["a", "b", "c", "d"]),
        # Interval boundaries must be positive
        (np.array([[-1, 2], [2, 3]]), ["a", "b"]),
        # Positive interval durations
        (np.array([[2, 1], [2, 3]]), ["a", "b"]),
        # Number of intervals must match number of labels
        (np.array([[2, 1], [2, 3]]), ["a"]),
        # Intervals must start at 0
        (np.array([[1, 2], [2, 3]]), ["a", "b"]),
    ],
)
def test_segment_structure_fail(metric, intervals, labels):
    metric(intervals, labels, intervals, labels)


@pytest.mark.xfail(raises=ValueError)
@pytest.mark.parametrize(
    "metric",
    [
        mir_eval.segment.pairwise,
        mir_eval.segment.rand_index,
        mir_eval.segment.ari,
        mir_eval.segment.mutual_information,
        mir_eval.segment.nce,
        mir_eval.segment.vmeasure,
    ],
)
def test_segment_structure_end_mismatch(metric):
    reference_intervals = np.array([[0, 1], [1, 2]])
    estimated_intervals = np.array([[0, 1], [1, 3]])
    labels = ["a", "b"]
    metric(reference_intervals, labels, estimated_intervals, labels)


@pytest.mark.parametrize(
    "metric",
    [
        mir_eval.segment.pairwise,
        mir_eval.segment.rand_index,
        mir_eval.segment.ari,
        mir_eval.segment.mutual_information,
        mir_eval.segment.nce,
        mir_eval.segment.vmeasure,
    ],
)
def test_segment_structure_perfect(metric):
    reference_intervals = np.array([[0, 1], [1, 2]])
    estimated_intervals = np.array([[0, 1], [1, 2]])
    labels = ["a", "b"]
    if metric == mir_eval.segment.mutual_information:
        assert np.allclose(
            metric(reference_intervals, labels, estimated_intervals, labels),
            [np.log(2), 1, 1],
        )
    else:
        assert np.allclose(
            metric(reference_intervals, labels, estimated_intervals, labels), 1
        )


@pytest.mark.parametrize("segment_data", file_sets, indirect=True)
def test_segment_functions(segment_data):
    ref_intervals, ref_labels, est_intervals, est_labels, expected_scores = segment_data

    # Compute scores
    scores = mir_eval.segment.evaluate(
        ref_intervals, ref_labels, est_intervals, est_labels
    )
    assert scores.keys() == expected_scores.keys()
    for metric in scores:
        assert np.allclose(scores[metric], expected_scores[metric], atol=A_TOL)


@pytest.mark.parametrize("segment_data", file_sets, indirect=True)
def test_segment_functions_permuted(segment_data):
    ref_intervals, ref_labels, est_intervals, est_labels, expected_scores = segment_data
    # Also check with permuted references
    idx = np.random.permutation(np.arange(len(ref_intervals)))

    perm_int = ref_intervals[idx]
    perm_lab = [ref_labels[_] for _ in idx]
    scores = mir_eval.segment.evaluate(perm_int, perm_lab, est_intervals, est_labels)
    assert scores.keys() == expected_scores.keys()
    for metric in scores:
        assert np.allclose(scores[metric], expected_scores[metric], atol=A_TOL)



"""
Unit tests for mir_eval.chord
"""

import mir_eval
import numpy as np
import pytest
import glob
import json

A_TOL = 1e-12

# Path to the fixture files
REF_GLOB = "data/chord/ref*.lab"
EST_GLOB = "data/chord/est*.lab"
SCORES_GLOB = "data/chord/output*.json"

ref_files = sorted(glob.glob(REF_GLOB))
est_files = sorted(glob.glob(EST_GLOB))
sco_files = sorted(glob.glob(SCORES_GLOB))

assert len(ref_files) == len(est_files) == len(sco_files) > 0

file_sets = list(zip(ref_files, est_files, sco_files))


@pytest.fixture
def chord_data(request):
    ref_f, est_f, sco_f = request.param
    with open(sco_f) as f:
        expected_scores = json.load(f)
    # Load in reference melody
    ref_intervals, ref_labels = mir_eval.io.load_labeled_intervals(ref_f)
    # Load in estimated melody
    est_intervals, est_labels = mir_eval.io.load_labeled_intervals(est_f)
    return ref_intervals, ref_labels, est_intervals, est_labels, expected_scores


@pytest.mark.parametrize(
    "pitch, semitone", [("Gbb", 5), ("G", 7), ("G#", 8), ("Cb", 11), ("B#", 0)]
)
def test_pitch_class_to_semitone_valid(pitch, semitone):
    assert mir_eval.chord.pitch_class_to_semitone(pitch) == semitone


@pytest.mark.parametrize("pitch", ["Cab", "#C", "bG"])
@pytest.mark.xfail(raises=mir_eval.chord.InvalidChordException)
def test_pitch_class_to_semitone_fail(pitch):
    mir_eval.chord.pitch_class_to_semitone(pitch)


@pytest.mark.parametrize(
    "degree, semitone",
    [
        ("b7", 10),
        ("#3", 5),
        ("1", 0),
        ("b1", -1),
        ("#7", 12),
        ("bb5", 5),
        ("11", 17),
        ("#13", 22),
    ],
)
def test_scale_degree_to_semitone(degree, semitone):
    assert mir_eval.chord.scale_degree_to_semitone(degree) == semitone


@pytest.mark.parametrize("degree", ["7b", "4#", "77", "15"])
@pytest.mark.xfail(raises=mir_eval.chord.InvalidChordException)
def test_scale_degree_to_semitone(degree):
    mir_eval.chord.scale_degree_to_semitone(degree)


@pytest.mark.parametrize(
    "degree, bitmap, modulo, length",
    [
        ("3", [0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0], True, 12),
        ("*3", [0, 0, 0, 0, -1, 0, 0, 0, 0, 0, 0, 0], True, 12),
        ("b1", [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1], True, 12),
        ("9", [0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0], True, 12),
        ("9", [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], False, 12),
        ("9", [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1], False, 15),
    ],
)
def test_scale_degree_to_bitmap(degree, bitmap, modulo, length):
    assert np.allclose(
        mir_eval.chord.scale_degree_to_bitmap(degree, modulo=modulo, length=length),
        bitmap,
    )


@pytest.mark.parametrize(
    "label", ["C", "Eb:min/5", "A#:dim7", "B:maj(*1,*5)/3", "A#:sus4", "A:(9,11)"]
)
def test_validate_chord_label(label):
    # For valid labels, calling the function without an error = pass
    mir_eval.chord.validate_chord_label(label)


@pytest.mark.parametrize(
    "label", ["C::maj", "C//5", "C((4)", "C5))", "C:maj(*3/3", "Cmaj*3/3)", "asdf"]
)
@pytest.mark.xfail(raises=mir_eval.chord.InvalidChordException)
def test_validate_bad_chord_label(label):
    mir_eval.chord.validate_chord_label(label)


@pytest.mark.parametrize(
    "label, split",
    [
        ("C", ["C", "maj", set(), "1"]),
        ("B:maj(*1,*3)/5", ["B", "maj", {"*1", "*3"}, "5"]),
        ("Ab:min/b3", ["Ab", "min", set(), "b3"]),
        ("N", ["N", "", set(), ""]),
        ("G:(3)", ["G", "", {"3"}, "1"]),
    ],
)
def test_split(label, split):
    assert mir_eval.chord.split(label) == split


@pytest.mark.parametrize(
    "label, split",
    [("C", ["C", "maj", set(), "1"]), ("C:minmaj7", ["C", "min", {"7"}, "1"])],
)
def test_split_extended(label, split):
    # Test with reducing extended chords
    mir_eval.chord.split(label, reduce_extended_chords=True) == split


@pytest.mark.xfail(raises=mir_eval.chord.InvalidChordException)
def test_split_fail():
    # Test that an exception is raised when a chord with an omission but no
    # quality is supplied
    mir_eval.chord.split("C(*5)")


# Arguments are root, quality, extensions, bass
@pytest.mark.parametrize(
    "label, split",
    [
        ("F#", ("F#", "", None, "")),
        ("F#:hdim7", ("F#", "hdim7", None, "")),
        ("F#:(*b3,4)", ("F#", "", ["*b3", "4"], "")),
        ("F#/b7", ("F#", "", None, "b7")),
        ("F#:(*b3,4)/b7", ("F#", "", ["*b3", "4"], "b7")),
        ("F#:hdim7/b7", ("F#", "hdim7", None, "b7")),
        ("F#:hdim7(*b3,4)/b7", ("F#", "hdim7", ["*b3", "4"], "b7")),
    ],
)
def test_join(label, split):
    # Test is relying on implicit parameter ordering here: root, quality, extensions, bass
    assert mir_eval.chord.join(*split) == label


@pytest.mark.parametrize(
    "bitmap, root, expected_bitmap",
    [
        ([1, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0], 0, [1, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0]),
        ([1, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0], 5, [1, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0]),
        (
            [1, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0],
            11,
            [0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 1],
        ),
    ],
)
def test_rotate_bitmaps_to_roots(bitmap, root, expected_bitmap):
    ans = mir_eval.chord.rotate_bitmaps_to_roots([bitmap], [root])
    assert np.all(ans == [expected_bitmap])


@pytest.mark.parametrize(
    "label, e_root, e_interval, e_bass, reduce, strict",
    [
        ("B:maj(*1,*3)/5", 11, [0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0], 7, False, False),
        ("G:dim", 7, [1, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0], 0, False, False),
        ("C:(3)/3", 0, [1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0], 4, False, False),
        ("A:9/b3", 9, [1, 0, 0, 1, 1, 0, 0, 1, 0, 0, 1, 0], 3, False, False),
        ("G:dim(4)/6", 7, [1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 0, 0], 9, False, False),
        ("A:9", 9, [1, 0, 1, 0, 1, 0, 0, 1, 0, 0, 1, 0], 0, True, False),
    ],
)
def test_chord_encode(label, e_root, e_interval, e_bass, reduce, strict):
    root, intervals, bass = mir_eval.chord.encode(
        label, reduce_extended_chords=reduce, strict_bass_intervals=strict
    )
    assert root == e_root, (root, e_root)
    assert np.all(intervals == e_interval), (intervals, e_interval)
    assert bass == e_bass, (bass, e_bass)


@pytest.mark.xfail(raises=mir_eval.chord.InvalidChordException)
def test_chord_encode_fail():
    # Non-chord bass notes *must* be explicitly named as extensions when
    #   strict_bass_intervals == True
    mir_eval.chord.encode(
        "G:dim(4)/6", reduce_extended_chords=False, strict_bass_intervals=True
    )


def test_encode_many():
    """Does all of the logic for checking encode_many"""
    labels = ["B:maj(*1,*3)/5", "B:maj(*1,*3)/5", "N", "C:min", "C:min"]
    expected_roots = [11, 11, -1, 0, 0]
    expected_intervals = [
        [0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0],
        [0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0],
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
        [1, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0],
        [1, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0],
    ]
    expected_basses = [7, 7, -1, 0, 0]

    roots, intervals, basses = mir_eval.chord.encode_many(labels)
    assert np.all(roots == expected_roots)
    assert np.all(intervals == expected_intervals)
    assert np.all(basses == expected_basses)


def __check_one_metric(metric, ref_label, est_label, score):
    """Checks that a metric function produces score given ref_label and
    est_label"""
    # We provide a dummy interval.  We're just checking one pair
    # of labels at a time.
    assert metric([ref_label], [est_label]) == score


def __check_not_comparable(metric, ref_label, est_label):
    """Checks that ref_label is not comparable to est_label by metric"""
    # Try to produce the warning
    with pytest.warns(
        UserWarning,
        match="No reference chords were comparable to estimated chords, returning 0.",
    ):
        score = mir_eval.chord.weighted_accuracy(
            metric([ref_label], [est_label]), np.array([1])
        )
        # And confirm that the metric is 0
        assert np.allclose(score, 0)


# TODO(ejhumphrey): Comparison functions lacking unit tests.
# test_root()


@pytest.mark.parametrize(
    "ref_label, est_label, score",
    [
        ("N", "N", 1.0),
        ("C:maj", "N", 0.0),
        ("C:maj", "C:aug", 0.0),
        ("C:maj", "C:dim", 0.0),
        ("C:min", "C:dim", 0.0),
        ("C:maj", "C:5", 0.0),
        ("C:maj", "C:sus4", 0.0),
        ("G:min", "G:sus2", 0.0),
        ("C:maj", "G:maj", 0.0),
        ("C:min", "C:hdim7", 0.0),
        ("C:min", "C:min7", 1.0),
        ("C:maj", "C:maj6", 1.0),
        ("F:maj", "F:min6", 0.0),
        ("C:maj7", "C:minmaj7", 1.0),
        ("A:maj", "A:7", 1.0),
        ("A:maj", "A:9", 1.0),
    ],
)
def test_mirex(ref_label, est_label, score):
    __check_one_metric(mir_eval.chord.mirex, ref_label, est_label, score)


@pytest.mark.parametrize("ref, est", [("C:5", "C:maj"), ("X", "N")])
def test_mirex_nocomp(ref, est):
    __check_not_comparable(mir_eval.chord.mirex, ref, est)


@pytest.mark.parametrize(
    "ref_label, est_label, score",
    [
        ("N", "N", 1.0),
        ("C:maj", "N", 0.0),
        ("C:maj", "C:aug", 1.0),
        ("C:maj", "C:dim", 0.0),
        ("C:min", "C:dim", 1.0),
        ("C:maj", "C:sus4", 1.0),
        ("G:min", "G:sus2", 0.0),
        ("C:maj", "G:maj", 0.0),
        ("C:min", "C:hdim7", 1.0),
        ("C:min", "C:min7", 1.0),
        ("C:maj", "C:maj6", 1.0),
        ("F:maj", "F:min6", 0.0),
        ("C:maj", "C:minmaj7", 0.0),
        ("A:maj", "A:7", 1.0),
        ("A:maj", "A:9", 1.0),
    ],
)
def test_thirds(ref_label, est_label, score):
    __check_one_metric(mir_eval.chord.thirds, ref_label, est_label, score)


def test_thirds_nocomp():
    __check_not_comparable(mir_eval.chord.thirds, "X", "N")


@pytest.mark.parametrize(
    "ref_label, est_label, score",
    [
        ("C:maj/5", "C:sus4/5", 1.0),
        ("G:min", "G:min/b3", 0.0),
        ("C:maj", "C:maj/5", 0.0),
        ("C:min/b3", "C:hdim7/b3", 1.0),
        ("C:min", "C:dim", 1.0),
    ],
)
def test_thirds_inv(ref_label, est_label, score):
    __check_one_metric(mir_eval.chord.thirds_inv, ref_label, est_label, score)


def test_thirds_inv_nocomp():
    __check_not_comparable(mir_eval.chord.thirds_inv, "X", "N")


@pytest.mark.parametrize(
    "ref_label, est_label, score",
    [
        ("C:min", "C:min7", 1.0),
        ("C:maj", "C:7", 1.0),
        ("C:maj", "C:aug", 0.0),
        ("C:min", "C:dim", 0.0),
        ("C:maj", "C:sus2", 0.0),
        ("C:maj", "C:sus4", 0.0),
        ("G:min", "G:minmaj7", 1.0),
        ("C:maj", "G:maj", 0.0),
        ("C:min", "C:hdim7", 0.0),
        ("C:min", "C:min6", 1.0),
    ],
)
def test_triads(ref_label, est_label, score):
    __check_one_metric(mir_eval.chord.triads, ref_label, est_label, score)


def test_triads_nocomp():
    __check_not_comparable(mir_eval.chord.triads, "X", "N")


@pytest.mark.parametrize(
    "ref_label, est_label, score",
    [
        ("C:maj/5", "C:maj7/5", 1.0),
        ("G:min", "G:min7/5", 0.0),
        ("C:maj", "C:7/5", 0.0),
        ("C:min/b3", "C:min6/b3", 1.0),
        ("C:min/b3", "C:dim/b3", 0.0),
    ],
)
def test_triads_inv(ref_label, est_label, score):
    __check_one_metric(mir_eval.chord.triads_inv, ref_label, est_label, score)


def test_triads_inv_nocomp():
    __check_not_comparable(mir_eval.chord.triads_inv, "X", "N")


@pytest.mark.parametrize(
    "ref_label, est_label, score",
    [
        ("C:min", "C:min7", 0.0),
        ("C:maj", "C:maj6", 0.0),
        ("C:7", "C:9", 1.0),
        ("C:maj7", "C:maj7/5", 1.0),
        ("C:sus2", "C:sus2/2", 1.0),
        ("C:7/3", "C:11/b7", 1.0),
        ("G:min", "G:sus2", 0.0),
        ("C:maj", "G:maj", 0.0),
        ("C:min", "C:hdim7", 0.0),
        ("C:min", "C:minmaj7", 0.0),
    ],
)
def test_tetrads(ref_label, est_label, score):
    __check_one_metric(mir_eval.chord.tetrads, ref_label, est_label, score)


def test_tetrads_nocomp():
    __check_not_comparable(mir_eval.chord.tetrads, "X", "N")


@pytest.mark.parametrize(
    "ref_label, est_label, score",
    [
        ("C:maj7/5", "C:maj7/3", 0.0),
        ("G:min", "G:min/b3", 0.0),
        ("C:7/5", "C:13/5", 1.0),
        ("C:min/b3", "C:hdim7/b3", 0.0),
        ("C:min9", "C:min7", 1.0),
    ],
)
def test_tetrads_inv(ref_label, est_label, score):
    __check_one_metric(mir_eval.chord.tetrads_inv, ref_label, est_label, score)


def test_tetrads_inv_nocomp():
    __check_not_comparable(mir_eval.chord.tetrads_inv, "X", "N")


@pytest.mark.parametrize(
    "ref_label, est_label, score",
    [
        ("N", "N", 1.0),
        ("C:maj", "N", 0.0),
        ("C:maj", "C:aug", 0.0),
        ("C:min", "C:dim", 0.0),
        ("G:maj7", "G", 1.0),
    ],
)
def test_majmin(ref_label, est_label, score):
    __check_one_metric(mir_eval.chord.majmin, ref_label, est_label, score)


@pytest.mark.parametrize("ref_label, est_label", [("C:aug", "C:maj"), ("X", "N")])
def test_majmin_nocomp(ref_label, est_label):
    __check_not_comparable(mir_eval.chord.majmin, ref_label, est_label)


@pytest.mark.parametrize(
    "ref_label, est_label, score",
    [
        ("C:maj/5", "C:sus4/5", 0.0),
        ("G:min", "G:min/b3", 0.0),
        ("C:maj/5", "C:maj/5", 1.0),
        ("C:min7", "C:min", 1.0),
        ("G:min/b3", "G:min/b3", 1.0),
        ("C:maj7/5", "C:maj/5", 1.0),
        ("C:7", "C:maj", 1.0),
    ],
)
def test_majmin_inv(ref_label, est_label, score):
    __check_one_metric(mir_eval.chord.majmin_inv, ref_label, est_label, score)


@pytest.mark.parametrize(
    "ref_label, est_label",
    [
        ("C:hdim7/b3", "C:min/b3"),
        ("C:maj/4", "C:maj/4"),
        ("C:maj/2", "C:sus2/2"),
        ("X", "N"),
    ],
)
def test_majmin_inv_nocomp(ref_label, est_label):
    __check_not_comparable(mir_eval.chord.majmin_inv, ref_label, est_label)


@pytest.mark.parametrize(
    "ref_label, est_label, score",
    [
        ("C:min", "C:min7", 0.0),
        ("C:maj", "C:maj6", 0.0),
        ("C:7", "C:9", 1.0),
        ("C:maj7", "C:maj7/5", 1.0),
        ("C:7/3", "C:11/b7", 1.0),
        ("G:min", "G:sus2", 0.0),
        ("C:maj", "G:maj", 0.0),
        ("C:7", "C:maj7", 0.0),
    ],
)
def test_sevenths(ref_label, est_label, score):
    __check_one_metric(mir_eval.chord.sevenths, ref_label, est_label, score)


@pytest.mark.parametrize(
    "ref_label, est_label", [("C:sus2", "C:sus2/2"), ("C:hdim7", "C:hdim7"), ("X", "N")]
)
def test_sevenths_nocomp(ref_label, est_label):
    __check_not_comparable(mir_eval.chord.sevenths, ref_label, est_label)


@pytest.mark.parametrize(
    "ref_label, est_label, score",
    [
        ("C:maj7/5", "C:maj7/3", 0.0),
        ("G:min", "G:min/b3", 0.0),
        ("C:7/5", "C:13/5", 1.0),
        ("C:min7/b7", "C:min7/b7", 1.0),
    ],
)
def test_sevenths_inv(ref_label, est_label, score):
    __check_one_metric(mir_eval.chord.sevenths_inv, ref_label, est_label, score)


@pytest.mark.parametrize(
    "ref_label, est_label", [("C:dim7/b3", "C:dim7/b3"), ("X", "N")]
)
def test_sevenths_inv_nocomp(ref_label, est_label):
    __check_not_comparable(mir_eval.chord.sevenths_inv, ref_label, est_label)


def test_directional_hamming_distance():
    ref_ivs = np.array([[0.0, 1.0], [1.0, 2.0], [2.0, 3.0]])
    est_ivs = np.array([[0.0, 0.9], [0.9, 1.8], [1.8, 2.5]])
    dhd_ref_to_est = (0.1 + 0.2 + 0.5) / 3.0
    dhd_est_to_ref = (0.0 + 0.1 + 0.2) / 2.5

    dhd = mir_eval.chord.directional_hamming_distance
    assert np.allclose(dhd_ref_to_est, dhd(ref_ivs, est_ivs))
    assert np.allclose(dhd_est_to_ref, dhd(est_ivs, ref_ivs))
    assert np.allclose(0, dhd(ref_ivs, ref_ivs))
    assert np.allclose(0, dhd(est_ivs, est_ivs))

    ivs_overlap_all = np.array([[0.0, 1.0], [0.9, 2.0]])
    ivs_overlap_one = np.array([[0.0, 1.0], [0.9, 2.0], [2.0, 3.0]])
    with pytest.raises(ValueError):
        dhd(ivs_overlap_all, est_ivs)
    with pytest.raises(ValueError):
        dhd(ivs_overlap_one, est_ivs)


def test_segmentation_functions():
    ref_ivs = np.array([[0.0, 2.0], [2.0, 2.5], [2.5, 3.2]])
    est_ivs = np.array([[0.0, 3.0], [3.0, 3.5]])
    true_oseg = 1.0 - 0.2 / 3.2
    true_useg = 1.0 - (1.0 + 0.2) / 3.5
    true_seg = min(true_oseg, true_useg)
    assert np.allclose(true_oseg, mir_eval.chord.overseg(ref_ivs, est_ivs))
    assert np.allclose(true_useg, mir_eval.chord.underseg(ref_ivs, est_ivs))
    assert np.allclose(true_seg, mir_eval.chord.seg(ref_ivs, est_ivs))

    ref_ivs = np.array([[0.0, 2.0], [2.0, 2.5], [2.5, 3.2]])
    est_ivs = np.array([[0.0, 2.0], [2.0, 2.5], [2.5, 3.2]])
    true_oseg = 1.0
    true_useg = 1.0
    true_seg = 1.0
    assert np.allclose(true_oseg, mir_eval.chord.overseg(ref_ivs, est_ivs))
    assert np.allclose(true_useg, mir_eval.chord.underseg(ref_ivs, est_ivs))
    assert np.allclose(true_seg, mir_eval.chord.seg(ref_ivs, est_ivs))

    ref_ivs = np.array([[0.0, 2.0], [2.0, 2.5], [2.5, 3.2]])
    est_ivs = np.array([[0.0, 3.2]])
    true_oseg = 1.0
    true_useg = 1 - 1.2 / 3.2
    true_seg = min(true_oseg, true_useg)
    assert np.allclose(true_oseg, mir_eval.chord.overseg(ref_ivs, est_ivs))
    assert np.allclose(true_useg, mir_eval.chord.underseg(ref_ivs, est_ivs))
    assert np.allclose(true_seg, mir_eval.chord.seg(ref_ivs, est_ivs))

    ref_ivs = np.array([[0.0, 2.0], [2.0, 2.5], [2.5, 3.2]])
    est_ivs = np.array([[3.2, 3.5]])
    true_oseg = 1.0
    true_useg = 1.0
    true_seg = 1.0
    assert np.allclose(true_oseg, mir_eval.chord.overseg(ref_ivs, est_ivs))
    assert np.allclose(true_useg, mir_eval.chord.underseg(ref_ivs, est_ivs))
    assert np.allclose(true_seg, mir_eval.chord.seg(ref_ivs, est_ivs))


def test_merge_chord_intervals():
    intervals = np.array([[0.0, 1.0], [1.0, 2.0], [2.0, 3], [3.0, 4.0], [4.0, 5.0]])
    labels = ["C:maj", "C:(1,3,5)", "A:maj", "A:maj7", "A:maj7/3"]
    assert np.allclose(
        np.array([[0.0, 2.0], [2.0, 3], [3.0, 4.0], [4.0, 5.0]]),
        mir_eval.chord.merge_chord_intervals(intervals, labels),
    )


def test_weighted_accuracy():
    # First, test for a warning on empty beats
    with pytest.warns(UserWarning, match="No nonzero weights, returning 0"):
        score = mir_eval.chord.weighted_accuracy(
            np.array([1, 0, 1]), np.array([0, 0, 0])
        )
        # And that the metric is 0
        assert np.allclose(score, 0)

    # len(comparisons) must equal len(weights)
    comparisons = np.array([1, 0, 1])
    weights = np.array([1, 1])

    with pytest.raises(ValueError):
        mir_eval.chord.weighted_accuracy(comparisons, weights)

    # Weights must all be positive
    comparisons = np.array([1, 1])
    weights = np.array([-1, -1])
    with pytest.raises(ValueError):
        mir_eval.chord.weighted_accuracy(comparisons, weights)

    # Make sure accuracy = 1 and 0 when all comparisons are True and False resp
    comparisons = np.array([1, 1, 1])
    weights = np.array([1, 1, 1])
    score = mir_eval.chord.weighted_accuracy(comparisons, weights)
    assert np.allclose(score, 1)
    comparisons = np.array([0, 0, 0])
    score = mir_eval.chord.weighted_accuracy(comparisons, weights)
    assert np.allclose(score, 0)


@pytest.mark.parametrize("chord_data", file_sets, indirect=True)
def test_chord_functions(chord_data):
    ref_intervals, ref_labels, est_intervals, est_labels, expected_scores = chord_data

    # Compute scores
    scores = mir_eval.chord.evaluate(
        ref_intervals, ref_labels, est_intervals, est_labels
    )
    # Compare them
    assert scores.keys() == expected_scores.keys()
    for metric in scores:
        assert np.allclose(scores[metric], expected_scores[metric], atol=A_TOL)


def test_quality_to_bitmap():
    # Test simple case
    assert np.all(
        mir_eval.chord.quality_to_bitmap("maj")
        == np.array([1, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0])
    )


@pytest.mark.xfail(raises=mir_eval.chord.InvalidChordException)
@pytest.mark.parametrize("quality", ["maj5", "2", "#7"])
def test_quality_to_bitmap_fail(quality):
    # Check exceptions for qualities not in the QUALITIES list
    mir_eval.chord.quality_to_bitmap(quality)


def test_validate():
    # Test that the validate function raises the appropriate errors and
    # warnings
    with pytest.warns() as w:
        # First, test for warnings on empty labels
        mir_eval.chord.validate([], [])
    assert len(w) == 2
    assert issubclass(w[-1].category, UserWarning)
    assert str(w[-1].message) == "Estimated labels are empty"
    assert issubclass(w[-2].category, UserWarning)
    assert str(w[-2].message) == "Reference labels are empty"

    # Test that error is thrown on different-length labels
    with pytest.raises(ValueError):
        mir_eval.chord.validate([], ["C"])


@pytest.mark.parametrize("quality", list(mir_eval.chord.QUALITIES.keys()))
def test_chord_qualities(quality):
    if len(quality) > 0:
        mir_eval.chord.encode(f"C:{quality}", reduce_extended_chords=True)
    else:
        mir_eval.chord.encode("C", reduce_extended_chords=True)



"""
unit tests for mir_eval.separation

load randomly generated source and estimated source signals and
the output from BSS_eval MATLAB implementation, make sure the results
from mir_eval numerically match.
"""

import numpy as np
import mir_eval
import glob
import pytest
import json
import os

A_TOL = 1e-2

REF_GLOB = "data/separation/ref*"
EST_GLOB = "data/separation/est*"
SCORES_GLOB = "data/separation/output*.json"

ref_files = sorted(glob.glob(REF_GLOB))
est_files = sorted(glob.glob(EST_GLOB))
sco_files = sorted(glob.glob(SCORES_GLOB))

assert len(ref_files) == len(est_files) == len(sco_files) > 0
file_sets = list(zip(ref_files, est_files, sco_files))

# Skip separation tests since deprecation
pytest.skip(allow_module_level=True)


@pytest.fixture
def separation_data(request):
    ref_f, est_f, sco_f = request.param
    with open(sco_f) as f:
        expected_results = json.load(f)
        expected_sources = expected_results["Sources"]
        expected_frames = expected_results["Framewise"]
        expected_images = expected_results["Images"]
        expected_image_frames = expected_results["Images Framewise"]

    # Load in example source separation data
    ref_sources = __load_and_stack_wavs(ref_f)
    est_sources = __load_and_stack_wavs(est_f)

    # Test inference for single source passed as single dimensional array
    if ref_sources.shape[0] == 1 and est_sources.shape[0] == 1:
        ref_sources = ref_sources[0]
        est_sources = est_sources[0]

    return (
        ref_sources,
        est_sources,
        expected_sources,
        expected_frames,
        expected_images,
        expected_image_frames,
    )


@pytest.fixture(autouse=True)
def seed_rng():
    # Seed the RNG before each test run
    np.random.seed(1999)


def __load_and_stack_wavs(directory):
    """Load all wavs in a directory and stack them vertically into a matrix"""
    stacked_audio_data = []
    global_fs = None
    for f in sorted(glob.glob(os.path.join(directory, "*.wav"))):
        audio_data, fs = mir_eval.io.load_wav(f)
        assert global_fs is None or fs == global_fs
        global_fs = fs
        stacked_audio_data.append(audio_data)
    return np.vstack(stacked_audio_data)


def __generate_multichannel(mono_sig, nchan=2, gain=1.0, reverse=False):
    """Turn a single channel (ie. mono) audio sample into a multichannel
    (e.g. stereo)
    Note: to achieve channels of silence pass gain=0
    """
    # add the channels dimension
    input_3d = np.atleast_3d(mono_sig)
    # get the desired number of channels
    stackin = [input_3d] * nchan
    # apply the gain to the new channels
    stackin[1:] = np.multiply(gain, stackin[1:])
    if reverse:
        # reverse the new channels
        stackin[1:] = stackin[1:][:][::-1]
    return np.dstack(stackin)


@pytest.mark.parametrize(
    "metric",
    [
        mir_eval.separation.bss_eval_sources,
        mir_eval.separation.bss_eval_sources_framewise,
        mir_eval.separation.bss_eval_images,
        mir_eval.separation.bss_eval_images_framewise,
    ],
)
def test_empty_input(metric):
    if (
        metric == mir_eval.separation.bss_eval_sources
        or metric == mir_eval.separation.bss_eval_images
    ):
        args = [np.array([]), np.array([])]
    elif (
        metric == mir_eval.separation.bss_eval_sources_framewise
        or metric == mir_eval.separation.bss_eval_images_framewise
    ):
        args = [np.array([]), np.array([]), 40, 20]

    with pytest.warns(UserWarning, match="is empty") as record:

        # First, test for a warning on empty audio data
        metric(*args)
        # And that the metric returns empty arrays
        assert np.allclose(metric(*args), np.array([]))

        # These warning counters are now offset by 1 because of the deprecation message
        assert "reference_sources is empty" in str(record[1].message)
        assert "estimated_sources is empty" in str(record[2].message)


@pytest.mark.parametrize(
    "metric",
    [
        mir_eval.separation.bss_eval_sources,
        mir_eval.separation.bss_eval_sources_framewise,
        mir_eval.separation.bss_eval_images,
        mir_eval.separation.bss_eval_images_framewise,
    ],
)
def test_silent_input(metric):
    # Test for error when there is a silent reference/estimated source
    if (
        metric == mir_eval.separation.bss_eval_images
        or metric == mir_eval.separation.bss_eval_images_framewise
    ):
        ref_sources = np.vstack(
            (np.zeros((1, 100, 2)), np.random.random_sample((2, 100, 2)))
        )
        est_sources = np.vstack(
            (np.zeros((1, 100, 2)), np.random.random_sample((2, 100, 2)))
        )
    else:
        ref_sources = np.vstack((np.zeros(100), np.random.random_sample((2, 100))))
        est_sources = np.vstack((np.zeros(100), np.random.random_sample((2, 100))))
    if (
        metric == mir_eval.separation.bss_eval_sources
        or metric == mir_eval.separation.bss_eval_images
    ):
        with pytest.raises(ValueError):
            metric(ref_sources[:2], est_sources[1:])
        with pytest.raises(ValueError):
            metric(ref_sources[1:], est_sources[:2])
    elif (
        metric == mir_eval.separation.bss_eval_sources_framewise
        or metric == mir_eval.separation.bss_eval_images_framewise
    ):
        with pytest.raises(ValueError):
            metric(ref_sources[:2], est_sources[1:], 40, 20)
        with pytest.raises(ValueError):
            metric(ref_sources[1:], est_sources[:2], 40, 20)


@pytest.mark.parametrize(
    "metric",
    [
        mir_eval.separation.bss_eval_sources_framewise,
        mir_eval.separation.bss_eval_images_framewise,
    ],
)
def test_partial_silence(metric):
    # Test for a full window of silence in reference/estimated source
    if metric == mir_eval.separation.bss_eval_sources_framewise:
        silence = np.zeros((2, 20))
        sound = np.random.random_sample((2, 20))
    elif metric == mir_eval.separation.bss_eval_images_framewise:
        silence = np.zeros((2, 20, 2))
        sound = np.random.random_sample((2, 20, 2))
    # test with silence in the reference
    results = metric(
        np.concatenate((sound, silence, sound), axis=1),
        np.concatenate((sound, sound, sound), axis=1),
        window=10,
        hop=10,
    )
    for measure in results:
        for idx, source in enumerate(measure):
            if idx < 2 or idx > 3:
                assert not np.isnan(source[idx])
            elif idx < 4:
                assert np.isnan(source[idx])
            else:
                raise ValueError("Testing error in partial silence test")
    # test with silence in the estimate
    results = metric(
        np.concatenate((sound, sound, sound), axis=1),
        np.concatenate((sound, silence, sound), axis=1),
        window=10,
        hop=10,
    )
    for measure in results:
        for idx, source in enumerate(measure):
            if idx < 2 or idx > 3:
                assert not np.isnan(source[idx])
            elif idx < 4:
                assert np.isnan(source[idx])
            else:
                raise ValueError("Testing error in partial silence test")


@pytest.mark.parametrize(
    "metric",
    [
        mir_eval.separation.bss_eval_sources,
        mir_eval.separation.bss_eval_sources_framewise,
        mir_eval.separation.bss_eval_images,
        mir_eval.separation.bss_eval_images_framewise,
    ],
)
def test_incompatible_shapes(metric):
    # Test for error when shape is different
    if (
        metric == mir_eval.separation.bss_eval_images
        or metric == mir_eval.separation.bss_eval_images_framewise
    ):
        sources_4 = np.random.random_sample((4, 100, 2))
        sources_3 = np.random.random_sample((3, 100, 2))
        sources_4_chan = np.random.random_sample((4, 100, 3))
    else:
        sources_4 = np.random.random_sample((4, 100))
        sources_3 = np.random.random_sample((3, 100))
    if (
        metric == mir_eval.separation.bss_eval_sources
        or metric == mir_eval.separation.bss_eval_images
    ):
        args1 = [sources_3, sources_4]
        args2 = [sources_4, sources_3]
    elif (
        metric == mir_eval.separation.bss_eval_sources_framewise
        or metric == mir_eval.separation.bss_eval_images_framewise
    ):
        args1 = [sources_3, sources_4, 40, 20]
        args2 = [sources_4, sources_3, 40, 20]
    with pytest.raises(ValueError):
        metric(*args1)
    with pytest.raises(ValueError):
        metric(*args2)
    if (
        metric == mir_eval.separation.bss_eval_images
        or metric == mir_eval.separation.bss_eval_images_framewise
    ):
        with pytest.raises(ValueError):
            metric(sources_4, sources_4_chan)


@pytest.mark.parametrize(
    "metric",
    [
        mir_eval.separation.bss_eval_sources,
        mir_eval.separation.bss_eval_sources_framewise,
        mir_eval.separation.bss_eval_images,
        mir_eval.separation.bss_eval_images_framewise,
    ],
)
def test_too_many_sources(metric):
    # Test for error when too many sources or references are provided
    many_sources = np.random.random_sample((mir_eval.separation.MAX_SOURCES * 2, 400))
    if metric == mir_eval.separation.bss_eval_sources:
        with pytest.raises(ValueError):
            metric(many_sources, many_sources)
    elif metric == mir_eval.separation.bss_eval_sources_framewise:
        with pytest.raises(ValueError):
            metric(many_sources, many_sources, 40, 20)


@pytest.mark.xfail(raises=ValueError)
@pytest.mark.parametrize(
    "metric",
    [
        mir_eval.separation.bss_eval_sources,
        mir_eval.separation.bss_eval_sources_framewise,
        mir_eval.separation.bss_eval_images,
        mir_eval.separation.bss_eval_images_framewise,
    ],
)
def test_too_many_dimensions(metric):
    # Test for detection of too high dimensioned images
    ref_sources = np.random.random_sample((4, 100, 2, 3))
    est_sources = np.random.random_sample((4, 100, 2, 3))
    metric(ref_sources, est_sources)


@pytest.mark.parametrize(
    "metric",
    [mir_eval.separation.bss_eval_sources, mir_eval.separation.bss_eval_images],
)
def test_default_permutation(metric):
    # Test for default permutation matrix when not computing permutation
    if metric == mir_eval.separation.bss_eval_sources:
        ref_sources = np.random.random_sample((4, 100))
        est_sources = np.random.random_sample((4, 100))
    elif metric == mir_eval.separation.bss_eval_images:
        ref_sources = np.random.random_sample((4, 100, 2))
        est_sources = np.random.random_sample((4, 100, 2))
    results = metric(ref_sources, est_sources, compute_permutation=False)
    assert np.array_equal(results[-1], np.asarray([0, 1, 2, 3]))


@pytest.mark.parametrize(
    "metric",
    [
        mir_eval.separation.bss_eval_sources_framewise,
        mir_eval.separation.bss_eval_images_framewise,
    ],
)
def test_framewise_small_window(metric):
    # Test for invalid win/hop parameter detection
    if metric == mir_eval.separation.bss_eval_sources_framewise:
        ref_sources = np.random.random_sample((4, 100))
        est_sources = np.random.random_sample((4, 100))
        comparison_fcn = mir_eval.separation.bss_eval_sources
    elif metric == mir_eval.separation.bss_eval_images_framewise:
        ref_sources = np.random.random_sample((4, 100, 2))
        est_sources = np.random.random_sample((4, 100, 2))
        comparison_fcn = mir_eval.separation.bss_eval_images

    # Test with window larger than source length
    assert np.allclose(
        np.squeeze(metric(ref_sources, est_sources, window=120, hop=20)),
        comparison_fcn(ref_sources, est_sources, False),
        atol=A_TOL,
    )
    # Test with hop larger than source length
    assert np.allclose(
        np.squeeze(metric(ref_sources, est_sources, window=20, hop=120)),
        comparison_fcn(ref_sources, est_sources, False),
        atol=A_TOL,
    )


@pytest.mark.parametrize("separation_data", file_sets, indirect=True)
def test_separation_functions(separation_data):
    (
        ref_sources,
        est_sources,
        expected_sources,
        expected_frames,
        expected_images,
        expected_image_frames,
    ) = separation_data

    # Compute scores
    scores = mir_eval.separation.evaluate(
        ref_sources,
        est_sources,
        window=expected_frames["win"],
        hop=expected_frames["hop"],
    )

    # Compare them
    for key in scores:
        if "Sources - " in key:
            test_data_name = key.replace("Sources - ", "")
            assert np.allclose(
                scores[key], expected_sources[test_data_name], atol=A_TOL
            )
        elif "Sources Frames - " in key:
            test_data_name = key.replace("Sources Frames - ", "")
            assert np.allclose(scores[key], expected_frames[test_data_name], atol=A_TOL)


@pytest.mark.parametrize("separation_data", file_sets, indirect=True)
def test_separation_images(separation_data):
    (
        ref_sources,
        est_sources,
        expected_sources,
        expected_frames,
        expected_images,
        expected_image_frames,
    ) = separation_data
    # Compute scores with images
    ref_images = __generate_multichannel(ref_sources, expected_images["nchan"])
    est_images = __generate_multichannel(
        est_sources,
        expected_images["nchan"],
        expected_images["gain"],
        expected_images["reverse"],
    )
    image_scores = mir_eval.separation.evaluate(ref_images, est_images)
    # Compare them
    for key in image_scores:
        if "Images - " in key:
            test_data_name = key.replace("Images - ", "")
            assert np.allclose(
                image_scores[key], expected_images[test_data_name], atol=A_TOL
            )


@pytest.mark.parametrize("separation_data", file_sets, indirect=True)
def test_separation_images_framewise(separation_data):
    (
        ref_sources,
        est_sources,
        expected_sources,
        expected_frames,
        expected_images,
        expected_image_frames,
    ) = separation_data

    # Compute scores with images framewise
    ref_images = __generate_multichannel(ref_sources, expected_image_frames["nchan"])
    est_images = __generate_multichannel(
        est_sources,
        expected_image_frames["nchan"],
        expected_image_frames["gain"],
        expected_image_frames["reverse"],
    )
    imageframe_scores = mir_eval.separation.evaluate(
        ref_images,
        est_images,
        window=expected_image_frames["win"],
        hop=expected_image_frames["hop"],
    )
    # Compare them
    for key in imageframe_scores:
        if "Images Frames - " in key:
            test_data_name = key.replace("Images Frames - ", "")
            assert np.allclose(
                imageframe_scores[key],
                expected_image_frames[test_data_name],
                atol=A_TOL,
            )

    # Catch a few exceptions in the evaluate function
    image_scores = mir_eval.separation.evaluate(ref_images, est_images)
    # make sure sources is not being evaluated on images
    assert "Sources - Source to Distortion" not in image_scores



"""
Unit tests for mir_eval.hierarchy
"""

import glob
import re

import json

import numpy as np
import scipy.sparse
import mir_eval
import pytest


A_TOL = 1e-12


@pytest.mark.parametrize("window", [5, 10, 15, 30, 90, None])
@pytest.mark.parametrize("frame_size", [0.1, 0.5, 1.0])
def test_tmeasure_pass(window, frame_size):
    # The estimate here gets none of the structure correct.
    ref = [[[0, 30]], [[0, 15], [15, 30]]]
    # convert to arrays
    ref = [np.asarray(_) for _ in ref]

    est = ref[:1]

    # The estimate should get 0 score here
    scores = mir_eval.hierarchy.tmeasure(ref, est, window=window, frame_size=frame_size)

    for k in scores:
        assert k == 0.0

    # The reference should get a perfect score here
    scores = mir_eval.hierarchy.tmeasure(ref, ref, window=window, frame_size=frame_size)

    for k in scores:
        assert k == 1.0


def test_tmeasure_warning():
    # Warn if there are missing boundaries from one layer to the next
    ref = [[[0, 5], [5, 10]], [[0, 10]]]

    ref = [np.asarray(_) for _ in ref]

    with pytest.warns(
        UserWarning, match="Segment hierarchy is inconsistent at level 1"
    ):
        mir_eval.hierarchy.tmeasure(ref, ref)


def test_tmeasure_fail_span():
    # Does not start at 0
    ref = [[[1, 10]], [[1, 5], [5, 10]]]

    ref = [np.asarray(_) for _ in ref]

    with pytest.raises(ValueError):
        mir_eval.hierarchy.tmeasure(ref, ref)

    # Does not end at the right time
    ref = [[[0, 5]], [[0, 5], [5, 6]]]
    ref = [np.asarray(_) for _ in ref]

    with pytest.raises(ValueError):
        mir_eval.hierarchy.tmeasure(ref, ref)

    # Two annotaions of different shape
    ref = [[[0, 10]], [[0, 5], [5, 10]]]
    ref = [np.asarray(_) for _ in ref]

    est = [[[0, 15]], [[0, 5], [5, 15]]]
    est = [np.asarray(_) for _ in est]

    with pytest.raises(ValueError):
        mir_eval.hierarchy.tmeasure(ref, est)


@pytest.mark.xfail(raises=ValueError)
@pytest.mark.parametrize(
    "window, frame_size",
    [(None, -1), (None, 0), (15, -1), (15, 0), (15, 30), (30, -1), (30, 0), (30, 60)],
)
def test_tmeasure_fail_frame_size(window, frame_size):
    ref = [[[0, 60]], [[0, 30], [30, 60]]]
    ref = [np.asarray(_) for _ in ref]

    mir_eval.hierarchy.tmeasure(ref, ref, window=window, frame_size=frame_size)


@pytest.mark.parametrize("frame_size", [0.1, 0.5, 1.0])
def test_lmeasure_pass(frame_size):
    # The estimate here gets none of the structure correct.
    ref = [[[0, 30]], [[0, 15], [15, 30]]]
    ref_lab = [["A"], ["a", "b"]]

    # convert to arrays
    ref = [np.asarray(_) for _ in ref]

    est = ref[:1]
    est_lab = ref_lab[:1]

    # The estimate should get 0 score here
    scores = mir_eval.hierarchy.lmeasure(
        ref, ref_lab, est, est_lab, frame_size=frame_size
    )

    for k in scores:
        assert k == 0.0

    # The reference should get a perfect score here
    scores = mir_eval.hierarchy.lmeasure(
        ref, ref_lab, ref, ref_lab, frame_size=frame_size
    )

    for k in scores:
        assert k == 1.0


def test_lmeasure_warning():
    # Warn if there are missing boundaries from one layer to the next
    ref = [[[0, 5], [5, 10]], [[0, 10]]]

    ref = [np.asarray(_) for _ in ref]
    ref_lab = [["a", "b"], ["A"]]

    with pytest.warns(
        UserWarning, match="Segment hierarchy is inconsistent at level 1"
    ):
        mir_eval.hierarchy.lmeasure(ref, ref_lab, ref, ref_lab)


def test_lmeasure_fail_span():
    # Does not start at 0
    ref = [[[1, 10]], [[1, 5], [5, 10]]]

    ref_lab = [["A"], ["a", "b"]]

    ref = [np.asarray(_) for _ in ref]

    with pytest.raises(ValueError):
        mir_eval.hierarchy.lmeasure(ref, ref_lab, ref, ref_lab)

    # Does not end at the right time
    ref = [[[0, 5]], [[0, 5], [5, 6]]]
    ref = [np.asarray(_) for _ in ref]

    with pytest.raises(ValueError):
        mir_eval.hierarchy.lmeasure(ref, ref_lab, ref, ref_lab)

    # Two annotations of different shape
    ref = [[[0, 10]], [[0, 5], [5, 10]]]
    ref = [np.asarray(_) for _ in ref]

    est = [[[0, 15]], [[0, 5], [5, 15]]]
    est = [np.asarray(_) for _ in est]

    with pytest.raises(ValueError):
        mir_eval.hierarchy.lmeasure(ref, ref_lab, est, ref_lab)


@pytest.mark.xfail(raises=ValueError)
@pytest.mark.parametrize("frame_size", [-1, 0])
def test_lmeasure_fail_frame_size(frame_size):
    ref = [[[0, 60]], [[0, 30], [30, 60]]]
    ref = [np.asarray(_) for _ in ref]
    ref_lab = [["A"], ["a", "b"]]

    mir_eval.hierarchy.lmeasure(ref, ref_lab, ref, ref_lab, frame_size=frame_size)


SCORES_GLOB = "data/hierarchy/output*.json"
sco_files = sorted(glob.glob(SCORES_GLOB))


@pytest.fixture
def hierarchy_outcomes(request):
    sco_f = request.param

    with open(sco_f) as fdesc:
        expected_scores = json.load(fdesc)
    window = float(re.match(r".*output_w=(\d+).json$", sco_f).groups()[0])

    return expected_scores, window


@pytest.mark.parametrize("hierarchy_outcomes", sco_files, indirect=True)
def test_hierarchy_regression(hierarchy_outcomes):
    expected_scores, window = hierarchy_outcomes

    # Hierarchy data is split across multiple lab files for these tests
    ref_files = sorted(glob.glob("data/hierarchy/ref*.lab"))
    est_files = sorted(glob.glob("data/hierarchy/est*.lab"))

    ref_hier = [mir_eval.io.load_labeled_intervals(_) for _ in ref_files]
    est_hier = [mir_eval.io.load_labeled_intervals(_) for _ in est_files]

    ref_ints = [seg[0] for seg in ref_hier]
    ref_labs = [seg[1] for seg in ref_hier]
    est_ints = [seg[0] for seg in est_hier]
    est_labs = [seg[1] for seg in est_hier]

    outputs = mir_eval.hierarchy.evaluate(
        ref_ints, ref_labs, est_ints, est_labs, window=window
    )

    assert outputs.keys() == expected_scores.keys()
    for key in expected_scores:
        assert np.allclose(expected_scores[key], outputs[key], atol=A_TOL)


def test_count_inversions():
    # inversion count = |{(i, j) : a[i] >= b[j]}|
    a = [2, 4, 6]
    b = [1, 2, 3, 4]

    # All inversions (a, b) are:
    # (2, 1), (2, 2)
    # (4, 1), (4, 2), (4, 3), (4, 4)
    # (6, 1), (6, 2), (6, 3), (6, 4)

    assert mir_eval.hierarchy._count_inversions(a, b) == 10

    # All inversions (b, a) are:
    # (2, 2)
    # (3, 2)
    # (4, 2), (4, 4)

    assert mir_eval.hierarchy._count_inversions(b, a) == 4

    # And test with repetitions
    a = [2, 2, 4]
    b = [1, 2, 4, 4]
    # counts: (a, b)
    # (2, 1), (2, 2)
    # (2, 1), (2, 2)
    # (4, 1), (4, 2), (4, 4), (4, 4)

    assert mir_eval.hierarchy._count_inversions(a, b) == 8

    # count: (b, a)
    # (2, 2), (2, 2)
    # (4, 2), (4, 2), (4, 4)
    # (4, 2), (4, 2), (4, 4)

    assert mir_eval.hierarchy._count_inversions(b, a) == 8


def test_meet():
    frame_size = 1
    int_hier = [
        np.array([[0, 10]]),
        np.array([[0, 6], [6, 10]]),
        np.array([[0, 2], [2, 4], [4, 6], [6, 8], [8, 10]]),
    ]

    lab_hier = [["X"], ["A", "B"], ["a", "b", "a", "c", "b"]]

    # Target output
    meet_truth = np.asarray(
        [
            [3, 3, 2, 2, 3, 3, 1, 1, 1, 1],  # (XAa)
            [3, 3, 2, 2, 3, 3, 1, 1, 1, 1],  # (XAa)
            [2, 2, 3, 3, 2, 2, 1, 1, 3, 3],  # (XAb)
            [2, 2, 3, 3, 2, 2, 1, 1, 3, 3],  # (XAb)
            [3, 3, 2, 2, 3, 3, 1, 1, 1, 1],  # (XAa)
            [3, 3, 2, 2, 3, 3, 1, 1, 1, 1],  # (XAa)
            [1, 1, 1, 1, 1, 1, 3, 3, 2, 2],  # (XBc)
            [1, 1, 1, 1, 1, 1, 3, 3, 2, 2],  # (XBc)
            [1, 1, 3, 3, 1, 1, 2, 2, 3, 3],  # (XBb)
            [1, 1, 3, 3, 1, 1, 2, 2, 3, 3],  # (XBb)
        ]
    )
    meet = mir_eval.hierarchy._meet(int_hier, lab_hier, frame_size)

    # Is it the right type?
    assert isinstance(meet, scipy.sparse.csr_matrix)
    meet = meet.toarray()

    # Does it have the right shape?
    assert meet.shape == (10, 10)

    # Does it have the right value?
    assert np.all(meet == meet_truth)


def test_compare_frame_rankings():
    # number of pairs (i, j)
    # where ref[i] < ref[j] and est[i] >= est[j]

    ref = np.asarray([1, 2, 3, 3])
    # ref pairs (transitive)
    # (1, 2), (1, 3), (1, 3), (2, 3), (2, 3)
    # ref pairs (non-transitive)
    # (1, 2), (2, 3), (2, 3)

    # Just count the normalizers
    # No self-inversions are possible from ref to itself
    inv, norm = mir_eval.hierarchy._compare_frame_rankings(ref, ref, transitive=True)
    assert inv == 0
    assert norm == 5.0

    inv, norm = mir_eval.hierarchy._compare_frame_rankings(ref, ref, transitive=False)
    assert inv == 0
    assert norm == 3.0

    est = np.asarray([1, 2, 1, 3])
    # In the transitive case, we lose two pairs
    # (1, 3) and (2, 2) -> (1, 1), (2, 1)
    inv, norm = mir_eval.hierarchy._compare_frame_rankings(ref, est, transitive=True)
    assert inv == 2
    assert norm == 5.0

    # In the non-transitive case, we only lose one pair
    # because (1,3) was not counted
    inv, norm = mir_eval.hierarchy._compare_frame_rankings(ref, est, transitive=False)
    assert inv == 1
    assert norm == 3.0

    # Do an all-zeros test
    ref = np.asarray([1, 1, 1, 1])
    inv, norm = mir_eval.hierarchy._compare_frame_rankings(ref, ref, transitive=True)
    assert inv == 0
    assert norm == 0.0



"""
Unit tests for mir_eval.onset
"""

import numpy as np
import pytest
import json
import mir_eval
import glob
import warnings

A_TOL = 1e-12

# Path to the fixture files
REF_GLOB = "data/onset/ref*.txt"
EST_GLOB = "data/onset/est*.txt"
SCORES_GLOB = "data/onset/output*.json"

ref_files = sorted(glob.glob(REF_GLOB))
est_files = sorted(glob.glob(EST_GLOB))
sco_files = sorted(glob.glob(SCORES_GLOB))

assert len(ref_files) == len(est_files) == len(sco_files) > 0

file_sets = list(zip(ref_files, est_files, sco_files))


@pytest.fixture
def onset_data(request):
    ref_f, est_f, sco_f = request.param
    with open(sco_f) as f:
        expected_scores = json.load(f)
    reference_onsets = mir_eval.io.load_events(ref_f)
    estimated_onsets = mir_eval.io.load_events(est_f)

    return reference_onsets, estimated_onsets, expected_scores


def test_onset_empty_warnings():
    with pytest.warns(UserWarning, match="Reference onsets are empty."):
        mir_eval.onset.f_measure(np.array([]), np.arange(10))

    with pytest.warns(UserWarning, match="Estimated onsets are empty."):
        mir_eval.onset.f_measure(np.arange(10), np.array([]))

    with pytest.warns(UserWarning, match="onsets are empty"):
        # Also verify that the score is 0
        assert np.allclose(mir_eval.onset.f_measure(np.array([]), np.array([])), 0)


@pytest.mark.xfail(raisses=ValueError)
@pytest.mark.parametrize(
    "onsets",
    [
        np.array([[1.0, 2.0]]),  # must be 1d ndarray
        np.array([1e10, 1e11]),  # must not be huge
        np.array([2.0, 1.0]),  # must be sorted
    ],
)
def test_onset_fail(onsets):
    mir_eval.onset.f_measure(onsets, onsets)


def test_onset_match():
    # Valid onsets which are the same produce a score of 1 for all metrics
    onsets = np.arange(10, dtype=np.float64)
    assert np.allclose(mir_eval.onset.f_measure(onsets, onsets), 1.0)


@pytest.mark.parametrize("onset_data", file_sets, indirect=True)
def test_onset_functions(onset_data):
    reference_onsets, estimated_onsets, expected_scores = onset_data

    # Compute scores
    scores = mir_eval.onset.evaluate(reference_onsets, estimated_onsets)
    # Compare them
    assert scores.keys() == expected_scores.keys()
    for metric in scores:
        assert np.allclose(scores[metric], expected_scores[metric], atol=A_TOL)



"""
Methods which sonify annotations for "evaluation by ear".
All functions return a raw signal at the specified sampling rate.
"""

import numpy as np
import scipy.signal
from numpy.lib.stride_tricks import as_strided
from scipy.interpolate import interp1d

from . import util
from . import chord


def clicks(times, fs, click=None, length=None):
    """Return a signal with the signal 'click' placed at each specified time

    Parameters
    ----------
    times : np.ndarray
        times to place clicks, in seconds
    fs : int
        desired sampling rate of the output signal
    click : np.ndarray
        click signal, defaults to a 1 kHz blip
    length : int
        desired number of samples in the output signal,
        defaults to ``times.max()*fs + click.shape[0] + 1``

    Returns
    -------
    click_signal : np.ndarray
        Synthesized click signal

    """
    # Create default click signal
    if click is None:
        # 1 kHz tone, 100ms
        click = np.sin(2 * np.pi * np.arange(fs * 0.1) * 1000 / (1.0 * fs))
        # Exponential decay
        click *= np.exp(-np.arange(fs * 0.1) / (fs * 0.01))
    # Set default length
    if length is None:
        length = int(times.max() * fs + click.shape[0] + 1)

    # Pre-allocate click signal
    click_signal = np.zeros(length)
    # Place clicks
    for time in times:
        # Compute the boundaries of the click
        start = int(time * fs)
        end = start + click.shape[0]
        # Make sure we don't try to output past the end of the signal
        if start >= length:
            break
        if end >= length:
            click_signal[start:] = click[: length - start]
            break
        # Normally, just add a click here
        click_signal[start:end] = click
    return click_signal


def time_frequency(
    gram, frequencies, times, fs, function=np.sin, length=None, n_dec=1, threshold=0.01
):
    r"""Reverse synthesis of a time-frequency representation of a signal

    Parameters
    ----------
    gram : np.ndarray
        ``gram[n, m]`` is the magnitude of ``frequencies[n]``
        from ``times[m]`` to ``times[m + 1]``

        Non-positive magnitudes are interpreted as silence.

    frequencies : np.ndarray
        array of size ``gram.shape[0]`` denoting the frequency (in Hz) of
        each row of gram

    times : np.ndarray, shape= ``(gram.shape[1],)`` or ``(gram.shape[1], 2)``
        Either the start time (in seconds) of each column in the gram,
        or the time interval (in seconds) corresponding to each column.

    fs : int
        desired sampling rate of the output signal

    function : function
        function to use to synthesize notes, should be :math:`2\pi`-periodic

    length : int
        desired number of samples in the output signal,
        defaults to ``times[-1]*fs``

    n_dec : int
        the number of decimals used to approximate each sonfied frequency.
        Defaults to 1 decimal place. Higher precision will be slower.

    threshold : float
        optimizes synthesis to only occur for frequencies that have a
        linear magnitude of at least one element in gram above the given threshold.

    Returns
    -------
    output : np.ndarray
        synthesized version of the piano roll

    """
    # Convert times to intervals if necessary
    time_converted = False
    if times.ndim == 1:
        # Convert to intervals
        times = np.hstack((times[:-1, np.newaxis], times[1:, np.newaxis]))
        # We'll need this to keep track of whether we should pad an interval on
        time_converted = True

    # Default value for length
    if length is None:
        length = int(np.max(times) * fs)

    last_time_in_secs = float(length) / fs

    if time_converted and times.shape[0] != gram.shape[1]:
        times = np.vstack((times, [np.max(times), last_time_in_secs]))

    if times.shape[0] != gram.shape[1]:
        raise ValueError(
            f"times.shape={times.shape} is incompatible with gram.shape={gram.shape}"
        )

    if frequencies.shape[0] != gram.shape[0]:
        raise ValueError(
            f"frequencies.shape={frequencies.shape} is incompatible with gram.shape={gram.shape}"
        )

    padding = [0, 0]
    stacking = []

    if times.min() > 0:
        # We need to pad a silence column on to gram at the beginning
        padding[0] = 1
        stacking.append([0, times.min()])

    stacking.append(times)

    if times.max() < last_time_in_secs:
        # We need to pad a silence column onto gram at the end
        padding[1] = 1
        stacking.append([times.max(), last_time_in_secs])

    gram = np.pad(gram, ((0, 0), padding), mode="constant")
    times = np.vstack(stacking)

    # Identify the time intervals that have some overlap with the duration
    idx = np.logical_and(times[:, 1] >= 0, times[:, 0] <= last_time_in_secs)
    gram = gram[:, idx]
    times = np.clip(times[idx], 0, last_time_in_secs)

    n_times = times.shape[0]

    # Threshold the tfgram to remove negative values
    gram = np.maximum(gram, 0)

    # Pre-allocate output signal
    output = np.zeros(length)
    if gram.shape[1] == 0:
        # There are no time intervals to process, so return
        # the empty signal.
        return output

    # Discard frequencies below threshold
    freq_keep = np.max(gram, axis=1) >= threshold

    gram = gram[freq_keep, :]
    frequencies = frequencies[freq_keep]

    # Interpolate the values in gram over the time grid.
    if n_times > 1:
        interpolator = interp1d(
            times[:, 0] * fs,
            gram[:, :n_times],
            kind="previous",
            bounds_error=False,
            fill_value=(gram[:, 0], gram[:, -1]),
        )
        signal = interpolator(np.arange(length))
    else:
        # NOTE: This is a special case where there is only one time interval.
        # scipy 1.10 and above handle this case directly with the interp1d above,
        # but older scipy's do not. This is a workaround for that.
        #
        # In the 0.9 release, we can bump the minimum scipy to 1.10 and remove this
        signal = np.tile(gram[:, 0], (1, length))

    for n, frequency in enumerate(frequencies):
        # Get a waveform of length samples at this frequency
        wave = _fast_synthesize(frequency, n_dec, fs, function, length)

        # Use a two-cycle ramp to smooth over transients
        period = 2 * int(fs / frequency)
        filter = np.ones(period) / period
        signal_n = scipy.signal.convolve(signal[n], filter, mode="same")

        # Mix the signal into the output
        output[:] += wave[: len(signal_n)] * signal_n

    # Normalize, but only if there's non-zero values
    norm = np.abs(output).max()
    if norm >= np.finfo(output.dtype).tiny:
        output /= norm

    return output


def _fast_synthesize(frequency, n_dec, fs, function, length):
    """Efficiently synthesize a signal.
    Generate one cycle, and simulate arbitrary repetitions
    using array indexing tricks.
    """
    # hack so that we can ensure an integer number of periods and samples
    # rounds frequency to 1st decimal, s.t. 10 * frequency will be an int
    frequency = np.round(frequency, n_dec)

    # Generate 10*frequency periods at this frequency
    # Equivalent to n_samples = int(n_periods * fs / frequency)
    # n_periods = 10*frequency is the smallest integer that guarantees
    # that n_samples will be an integer, since assuming 10*frequency
    # is an integer
    n_samples = int(10.0**n_dec * fs)

    short_signal = function(2.0 * np.pi * np.arange(n_samples) * frequency / fs)

    # Calculate the number of loops we need to fill the duration
    n_repeats = int(np.ceil(length / float(short_signal.shape[0])))

    # Simulate tiling the short buffer by using stride tricks
    long_signal = as_strided(
        short_signal,
        shape=(n_repeats, len(short_signal)),
        strides=(0, short_signal.itemsize),
    )

    # Use a flatiter to simulate a long 1D buffer
    return long_signal.flat


def pitch_contour(
    times, frequencies, fs, amplitudes=None, function=np.sin, length=None, kind="linear"
):
    r"""Sonify a pitch contour.

    Parameters
    ----------
    times : np.ndarray
        time indices for each frequency measurement, in seconds
    frequencies : np.ndarray
        frequency measurements, in Hz.
        Non-positive measurements or NaNs will be interpreted as un-voiced samples.
    fs : int
        desired sampling rate of the output signal
    amplitudes : np.ndarray
        amplitude measurements, nonnegative
        defaults to ``np.ones((length,))``
    function : function
        function to use to synthesize notes, should be :math:`2\pi`-periodic
    length : int
        desired number of samples in the output signal,
        defaults to ``max(times)*fs``
    kind : str
        Interpolation mode for the frequency and amplitude values.
        See: ``scipy.interpolate.interp1d`` for valid settings.

    Returns
    -------
    output : np.ndarray
        synthesized version of the pitch contour
    """
    fs = float(fs)

    if length is None:
        length = int(times.max() * fs)

    # Squash the negative frequencies.
    # wave(0) = 0, so clipping here will un-voice the corresponding instants
    frequencies = np.maximum(frequencies, 0.0)
    # Convert nans to zeros to unvoice
    frequencies = np.nan_to_num(frequencies, copy=False)

    # Build a frequency interpolator
    f_interp = interp1d(
        times * fs,
        2 * np.pi * frequencies / fs,
        kind=kind,
        fill_value=0.0,
        bounds_error=False,
        copy=False,
    )

    # Estimate frequency at sample points
    f_est = f_interp(np.arange(length))

    if amplitudes is None:
        a_est = np.ones((length,))
    else:
        # build an amplitude interpolator
        a_interp = interp1d(
            times * fs,
            amplitudes,
            kind=kind,
            fill_value=0.0,
            bounds_error=False,
            copy=False,
        )
        a_est = a_interp(np.arange(length))

    # Sonify the waveform
    return a_est * function(np.cumsum(f_est))


def chroma(chromagram, times, fs, **kwargs):
    """Reverse synthesis of a chromagram (semitone matrix)

    Parameters
    ----------
    chromagram : np.ndarray, shape=(12, times.shape[0])
        Chromagram matrix, where each row represents a semitone [C->Bb]
        i.e., ``chromagram[3, j]`` is the magnitude of D# from ``times[j]`` to
        ``times[j + 1]``
    times : np.ndarray, shape=(len(chord_labels),) or (len(chord_labels), 2)
        Either the start time of each column in the chromagram,
        or the time interval corresponding to each column.
    fs : int
        Sampling rate to synthesize audio data at
    **kwargs
        Additional keyword arguments to pass to
        :func:`mir_eval.sonify.time_frequency`

    Returns
    -------
    output : np.ndarray
        Synthesized chromagram

    """
    # We'll just use time_frequency with a Shepard tone-gram
    # To create the Shepard tone-gram, we copy the chromagram across 7 octaves
    n_octaves = 7
    # starting from C2
    base_note = 24
    # and weight each octave by a normal distribution
    # The normal distribution has mean 72 (one octave above middle C)
    # and std 6 (one half octave)
    mean = 72
    std = 6
    notes = np.arange(12 * n_octaves) + base_note
    shepard_weight = np.exp(-((notes - mean) ** 2.0) / (2.0 * std**2.0))
    # Copy the chromagram matrix vertically n_octaves times
    gram = np.tile(chromagram.T, n_octaves).T
    # This fixes issues if the supplied chromagram is int type
    gram = gram.astype(float)
    # Apply Sheppard weighting
    gram *= shepard_weight.reshape(-1, 1)
    # Compute frequencies
    frequencies = 440.0 * (2.0 ** ((notes - 69) / 12.0))
    return time_frequency(gram, frequencies, times, fs, **kwargs)


def chords(chord_labels, intervals, fs, **kwargs):
    """Synthesizes chord labels

    Parameters
    ----------
    chord_labels : list of str
        List of chord label strings.
    intervals : np.ndarray, shape=(len(chord_labels), 2)
        Start and end times of each chord label
    fs : int
        Sampling rate to synthesize at
    **kwargs
        Additional keyword arguments to pass to
        :func:`mir_eval.sonify.time_frequency`

    Returns
    -------
    output : np.ndarray
        Synthesized chord labels

    """
    util.validate_intervals(intervals)

    # Convert from labels to chroma
    roots, interval_bitmaps, _ = chord.encode_many(chord_labels)
    chromagram = np.array(
        [
            np.roll(interval_bitmap, root)
            for (interval_bitmap, root) in zip(interval_bitmaps, roots)
        ]
    ).T

    return chroma(chromagram, intervals, fs, **kwargs)



"""Unit tests for utils"""

import collections

import pytest
import numpy as np
import mir_eval
from mir_eval import util


def test_interpolate_intervals():
    """Check that an interval set is interpolated properly, with boundaries
    conditions and out-of-range values.
    """
    labels = list("abc")
    intervals = np.array([(n, n + 1.0) for n in range(len(labels))])
    time_points = [-1.0, 0.1, 0.9, 1.0, 2.3, 4.0]
    expected_ans = ["N", "a", "a", "b", "c", "N"]
    assert (
        util.interpolate_intervals(intervals, labels, time_points, "N") == expected_ans
    )


def test_interpolate_intervals_gap():
    """Check that an interval set is interpolated properly, with gaps."""
    labels = list("abc")
    intervals = np.array([[0.5, 1.0], [1.5, 2.0], [2.5, 3.0]])
    time_points = [0.0, 0.75, 1.25, 1.75, 2.25, 2.75, 3.5]
    expected_ans = ["N", "a", "N", "b", "N", "c", "N"]
    assert (
        util.interpolate_intervals(intervals, labels, time_points, "N") == expected_ans
    )


@pytest.mark.xfail(raises=ValueError)
def test_interpolate_intervals_badtime():
    """Check that interpolate_intervals throws an exception if
    input is unordered.
    """
    labels = list("abc")
    intervals = np.array([(n, n + 1.0) for n in range(len(labels))])
    time_points = [-1.0, 0.1, 0.9, 0.8, 2.3, 4.0]
    mir_eval.util.interpolate_intervals(intervals, labels, time_points)


def test_intervals_to_samples():
    """Check that an interval set is sampled properly, with boundaries
    conditions and out-of-range values.
    """
    labels = list("abc")
    intervals = np.array([(n, n + 1.0) for n in range(len(labels))])

    expected_times = [0.0, 0.5, 1.0, 1.5, 2.0, 2.5]
    expected_labels = ["a", "a", "b", "b", "c", "c"]
    result = util.intervals_to_samples(
        intervals, labels, offset=0, sample_size=0.5, fill_value="N"
    )
    assert result[0] == expected_times
    assert result[1] == expected_labels

    expected_times = [0.25, 0.75, 1.25, 1.75, 2.25, 2.75]
    expected_labels = ["a", "a", "b", "b", "c", "c"]
    result = util.intervals_to_samples(
        intervals, labels, offset=0.25, sample_size=0.5, fill_value="N"
    )
    assert result[0] == expected_times
    assert result[1] == expected_labels


def test_intersect_files():
    """Check that two non-identical produce correct results."""
    flist1 = ["/a/b/abc.lab", "/c/d/123.lab", "/e/f/xyz.lab"]
    flist2 = ["/g/h/xyz.npy", "/i/j/123.txt", "/k/l/456.lab"]
    sublist1, sublist2 = util.intersect_files(flist1, flist2)
    assert sublist1 == ["/e/f/xyz.lab", "/c/d/123.lab"]
    assert sublist2 == ["/g/h/xyz.npy", "/i/j/123.txt"]
    sublist1, sublist2 = util.intersect_files(flist1[:1], flist2[:1])
    assert sublist1 == []
    assert sublist2 == []


def test_merge_labeled_intervals():
    """Check that two labeled interval sequences merge correctly."""
    x_intvs = np.array([[0.0, 0.44], [0.44, 2.537], [2.537, 4.511], [4.511, 6.409]])
    x_labels = ["A", "B", "C", "D"]
    y_intvs = np.array([[0.0, 0.464], [0.464, 2.415], [2.415, 4.737], [4.737, 6.409]])
    y_labels = [0, 1, 2, 3]
    expected_intvs = [
        [0.0, 0.44],
        [0.44, 0.464],
        [0.464, 2.415],
        [2.415, 2.537],
        [2.537, 4.511],
        [4.511, 4.737],
        [4.737, 6.409],
    ]
    expected_x_labels = ["A", "B", "B", "B", "C", "D", "D"]
    expected_y_labels = [0, 0, 1, 2, 2, 2, 3]
    new_intvs, new_x_labels, new_y_labels = util.merge_labeled_intervals(
        x_intvs, x_labels, y_intvs, y_labels
    )

    assert new_x_labels == expected_x_labels
    assert new_y_labels == expected_y_labels
    assert new_intvs.tolist() == expected_intvs

    # Check that invalid inputs raise a ValueError
    y_intvs[-1, -1] = 10.0
    with pytest.raises(ValueError):
        util.merge_labeled_intervals(x_intvs, x_labels, y_intvs, y_labels)


def test_boundaries_to_intervals():
    # Basic tests
    boundaries = np.arange(10)
    correct_intervals = np.array([np.arange(10 - 1), np.arange(1, 10)]).T
    intervals = mir_eval.util.boundaries_to_intervals(boundaries)
    assert np.all(intervals == correct_intervals)


def test_adjust_events():
    # Test appending at the end
    events = np.arange(1, 11)
    labels = [str(n) for n in range(10)]
    new_e, new_l = mir_eval.util.adjust_events(events, labels, 0.0, 11.0)
    assert new_e[0] == 0.0
    assert new_l[0] == "__T_MIN"
    assert new_e[-1] == 11.0
    assert new_l[-1] == "__T_MAX"
    assert np.all(new_e[1:-1] == events)
    assert new_l[1:-1] == labels

    # Test trimming
    new_e, new_l = mir_eval.util.adjust_events(events, labels, 0.0, 9.0)
    assert new_e[0] == 0.0
    assert new_l[0] == "__T_MIN"
    assert new_e[-1] == 9.0
    assert np.all(new_e[1:] == events[:-1])
    assert new_l[1:] == labels[:-1]


def test_bipartite_match():
    # This test constructs a graph as follows:
    #   v9 -- (u0)
    #   v8 -- (u0, u1)
    #   v7 -- (u0, u1, u2)
    #   ...
    #   v0 -- (u0, u1, ..., u9)
    #
    # This structure and ordering of this graph should force Hopcroft-Karp to
    # hit each algorithm/layering phase
    #
    G = collections.defaultdict(list)

    u_set = [f"u{_:d}" for _ in range(10)]
    v_set = [f"v{_:d}" for _ in range(len(u_set) + 1)]
    for i, u in enumerate(u_set):
        for v in v_set[: -i - 1]:
            G[v].append(u)

    matching = util._bipartite_match(G)

    # Make sure that each u vertex is matched
    assert len(matching) == len(u_set)

    # Make sure that there are no duplicate keys
    lhs = {k for k in matching}
    rhs = {matching[k] for k in matching}

    assert len(matching) == len(lhs)
    assert len(matching) == len(rhs)

    # Finally, make sure that all detected edges are present in G
    for k in matching:
        v = matching[k]
        assert v in G[k] or k in G[v]


def test_outer_distance_mod_n():
    ref = [1.0, 2.0, 3.0]
    est = [1.1, 6.0, 1.9, 5.0, 10.0]
    expected = np.array(
        [
            [0.1, 5.0, 0.9, 4.0, 3.0],
            [0.9, 4.0, 0.1, 3.0, 4.0],
            [1.9, 3.0, 1.1, 2.0, 5.0],
        ]
    )
    actual = mir_eval.util._outer_distance_mod_n(ref, est)
    assert np.allclose(actual, expected)

    ref = [13.0, 14.0, 15.0]
    est = [1.1, 6.0, 1.9, 5.0, 10.0]
    expected = np.array(
        [
            [0.1, 5.0, 0.9, 4.0, 3.0],
            [0.9, 4.0, 0.1, 3.0, 4.0],
            [1.9, 3.0, 1.1, 2.0, 5.0],
        ]
    )
    actual = mir_eval.util._outer_distance_mod_n(ref, est)
    assert np.allclose(actual, expected)


def test_match_events():
    ref = [1.0, 2.0, 3.0]
    est = [1.1, 6.0, 1.9, 5.0, 10.0]
    expected = [(0, 0), (1, 2)]
    actual = mir_eval.util.match_events(ref, est, 0.5)
    assert actual == expected

    ref = [1.0, 2.0, 3.0, 11.9]
    est = [1.1, 6.0, 1.9, 5.0, 10.0, 0.0]
    expected = [(0, 0), (1, 2), (3, 5)]
    actual = mir_eval.util.match_events(
        ref, est, 0.5, distance=mir_eval.util._outer_distance_mod_n
    )
    assert actual == expected


def test_fast_hit_windows():
    ref = [1.0, 2.0, 3.0]
    est = [1.1, 6.0, 1.9, 5.0, 10.0]

    ref_fast, est_fast = mir_eval.util._fast_hit_windows(ref, est, 0.5)
    ref_slow, est_slow = np.where(np.abs(np.subtract.outer(ref, est)) <= 0.5)

    assert np.all(ref_fast == ref_slow)
    assert np.all(est_fast == est_slow)


@pytest.mark.xfail(raises=ValueError)
@pytest.mark.parametrize(
    "intervals",
    [
        # Test for ValueError when interval shape is invalid
        np.array([[1.0], [2.5], [5.0]]),
        # Test for ValueError when times are negative
        np.array([[1.0, -2.0], [2.5, 3.0], [5.0, 6.0]]),
        # Test for ValueError when duration is zero
        np.array([[1.0, 2.0], [2.5, 2.5], [5.0, 6.0]]),
        # Test for ValueError when duration is negative
        np.array([[1.0, 2.0], [2.5, 1.5], [5.0, 6.0]]),
    ],
)
def test_validate_intervals(intervals):
    mir_eval.util.validate_intervals(intervals)


@pytest.mark.xfail(raises=ValueError)
@pytest.mark.parametrize(
    "events",
    [
        # Test for ValueError when max_time is violated
        np.array([100.0, 100000.0]),
        # Test for ValueError when events aren't 1-d arrays
        np.array([[1.0, 2.0], [3.0, 4.0]]),
        # Test for ValueError when event times are not increasing
        np.array([1.0, 2.0, 5.0, 3.0]),
    ],
)
def test_validate_events(events):
    mir_eval.util.validate_events(events)


@pytest.mark.xfail(raises=ValueError)
@pytest.mark.parametrize(
    "freqs",
    [
        # Test for ValueError when max_freq is violated
        np.array([100, 10000]),
        # Test for ValueError when min_freq is violated
        np.array([2, 200]),
        # Test for ValueError when events aren't 1-d arrays
        np.array([[100, 200], [300, 400]]),
        # Test for ValueError when allow_negatives is false and negative values
        # are passed
        np.array([-100, 200]),
    ],
)
def test_validate_frequencies(freqs):
    mir_eval.util.validate_frequencies(freqs, 5000, 20, allow_negatives=False)


@pytest.mark.xfail(raises=ValueError)
@pytest.mark.parametrize(
    "freqs",
    [
        # Test for ValueError when max_freq is violated and allow_negatives=True
        np.array([100, -100000]),
        # Test for ValueError when min_freq is violated and allow_negatives=True
        np.array([-2, 200]),
    ],
)
def test_validate_frequencies_negative(freqs):
    mir_eval.util.validate_frequencies(freqs, 5000, 20, allow_negatives=True)


def test_has_kwargs():
    def __test(target, f):
        assert target == mir_eval.util.has_kwargs(f)

    def f1(_):
        return None

    def f2(_=5):
        return None

    def f3(*_):
        return None

    def f4(_, **kw):
        return None

    def f5(_=5, **kw):
        return None

    assert not mir_eval.util.has_kwargs(f1)
    assert not mir_eval.util.has_kwargs(f2)
    assert not mir_eval.util.has_kwargs(f3)
    assert mir_eval.util.has_kwargs(f4)
    assert mir_eval.util.has_kwargs(f5)


@pytest.mark.parametrize(
    "x,labels,x_true,lab_true",
    [
        (
            np.asarray([[10, 20], [0, 10]]),
            ["a", "b"],
            np.asarray([[0, 10], [10, 20]]),
            ["b", "a"],
        ),
        (
            np.asarray([[0, 10], [10, 20]]),
            ["b", "a"],
            np.asarray([[0, 10], [10, 20]]),
            ["b", "a"],
        ),
    ],
)
def test_sort_labeled_intervals_with_labels(x, labels, x_true, lab_true):
    xs, ls = mir_eval.util.sort_labeled_intervals(x, labels)
    assert np.allclose(xs, x_true)
    assert ls == lab_true


@pytest.mark.parametrize(
    "x,x_true",
    [
        (np.asarray([[10, 20], [0, 10]]), np.asarray([[0, 10], [10, 20]])),
        (np.asarray([[0, 10], [10, 20]]), np.asarray([[0, 10], [10, 20]])),
    ],
)
def test_sort_labeled_intervals_without_labels(x, x_true):
    xs = mir_eval.util.sort_labeled_intervals(x)
    assert np.allclose(xs, x_true)



"""
Tests for mir_eval.key
"""

import mir_eval
import pytest
import glob
import json
import numpy as np

A_TOL = 1e-12

# Path to the fixture files
REF_GLOB = "data/key/ref*.txt"
EST_GLOB = "data/key/est*.txt"
SCORES_GLOB = "data/key/output*.json"

ref_files = sorted(glob.glob(REF_GLOB))
est_files = sorted(glob.glob(EST_GLOB))
sco_files = sorted(glob.glob(SCORES_GLOB))

assert len(ref_files) == len(est_files) == len(sco_files) > 0
file_sets = list(zip(ref_files, est_files, sco_files))


@pytest.mark.parametrize(
    "good_key",
    ["C major", "c major", "C# major", "Bb minor", "db minor", "X", "x", "C other"],
)
@pytest.mark.parametrize(
    "bad_key", ["C maj", "Cb major", "C", "K major", "F## minor" "X other", "x minor"]
)
def test_key_function_fail(good_key, bad_key):
    score = mir_eval.key.weighted_score(good_key, good_key)
    assert score == 1.0

    with pytest.raises(ValueError):
        mir_eval.key.weighted_score(good_key, bad_key)
    with pytest.raises(ValueError):
        mir_eval.key.weighted_score(bad_key, good_key)


@pytest.fixture
def key_data(request):
    ref_f, est_f, sco_f = request.param
    with open(sco_f) as f:
        expected_scores = json.load(f)
    reference_key = mir_eval.io.load_key(ref_f)
    estimated_key = mir_eval.io.load_key(est_f)

    return reference_key, estimated_key, expected_scores


@pytest.mark.parametrize("key_data", file_sets, indirect=True)
def test_key_functions(key_data):
    reference_key, estimated_key, expected_scores = key_data
    # Compute scores
    scores = mir_eval.key.evaluate(reference_key, estimated_key)
    # Compare them
    assert scores.keys() == expected_scores.keys()
    for metric in scores:
        assert np.allclose(scores[metric], expected_scores[metric], atol=A_TOL)



"""
Some unit tests for the pattern discovery task.
"""

import numpy as np
import json
import mir_eval
import glob
import pytest

A_TOL = 1e-12

# Path to the fixture files
REF_GLOB = "data/pattern/ref*.txt"
EST_GLOB = "data/pattern/est*.txt"
SCORES_GLOB = "data/pattern/output*.json"

ref_files = sorted(glob.glob(REF_GLOB))
est_files = sorted(glob.glob(EST_GLOB))
sco_files = sorted(glob.glob(SCORES_GLOB))

assert len(ref_files) == len(est_files) == len(sco_files) > 0

file_sets = list(zip(ref_files, est_files, sco_files))


@pytest.fixture
def pattern_data(request):
    ref_f, est_f, sco_f = request.param
    with open(sco_f) as f:
        expected_scores = json.load(f)
    reference_patterns = mir_eval.io.load_patterns(ref_f)
    estimated_patterns = mir_eval.io.load_patterns(est_f)

    return reference_patterns, estimated_patterns, expected_scores


@pytest.mark.parametrize(
    "metric",
    [
        mir_eval.pattern.standard_FPR,
        mir_eval.pattern.establishment_FPR,
        mir_eval.pattern.occurrence_FPR,
        mir_eval.pattern.three_layer_FPR,
        mir_eval.pattern.first_n_three_layer_P,
        mir_eval.pattern.first_n_target_proportion_R,
    ],
)
def test_pattern_empty(metric):
    # First, test for a warning on empty pattern
    with pytest.warns(UserWarning, match="Reference patterns are empty"):
        metric([[[]]], [[[(100, 20)]]])

    with pytest.warns(UserWarning, match="Estimated patterns are empty"):
        metric([[[(100, 20)]]], [[[]]])

    with pytest.warns(UserWarning, match="patterns are empty"):
        # And that the metric is 0
        assert np.allclose(metric([[[]]], [[[]]]), 0)


@pytest.mark.parametrize(
    "metric",
    [
        mir_eval.pattern.standard_FPR,
        mir_eval.pattern.establishment_FPR,
        mir_eval.pattern.occurrence_FPR,
        mir_eval.pattern.three_layer_FPR,
        mir_eval.pattern.first_n_three_layer_P,
        mir_eval.pattern.first_n_target_proportion_R,
    ],
)
@pytest.mark.parametrize(
    "patterns",
    [
        [[[(100, 20)]], []],  # patterns must have at least one occurrence
        [[[(100, 20, 3)]]],  # (onset, midi) tuple must contain 2 elements
    ],
)
@pytest.mark.xfail(raises=ValueError)
def test_pattern_failure(metric, patterns):
    metric(patterns, patterns)


@pytest.mark.parametrize(
    "metric",
    [
        mir_eval.pattern.standard_FPR,
        mir_eval.pattern.establishment_FPR,
        mir_eval.pattern.occurrence_FPR,
        mir_eval.pattern.three_layer_FPR,
        mir_eval.pattern.first_n_three_layer_P,
        mir_eval.pattern.first_n_target_proportion_R,
    ],
)
def test_pattern_perfect(metric):
    # Valid patterns which are the same produce a score of 1 for all metrics
    patterns = [[[(100, 20), (200, 30)]]]
    assert np.allclose(metric(patterns, patterns), 1)


@pytest.mark.parametrize("pattern_data", file_sets, indirect=True)
def test_pattern_functions(pattern_data):
    reference_patterns, estimated_patterns, expected_scores = pattern_data
    # Compute scores
    scores = mir_eval.pattern.evaluate(reference_patterns, estimated_patterns)
    # Compare them
    assert scores.keys() == expected_scores.keys()
    for metric in scores:
        assert np.allclose(scores[metric], expected_scores[metric], atol=A_TOL)



# CREATED: 2/9/16 2:27 PM by Justin Salamon <justin.salamon@nyu.edu>

import mir_eval
import numpy as np
import glob
import json
import pytest

A_TOL = 1e-12

# Path to the fixture files
REF_GLOB = "data/transcription/ref*.txt"
EST_GLOB = "data/transcription/est*.txt"
SCORES_GLOB = "data/transcription/output*.json"

ref_files = sorted(glob.glob(REF_GLOB))
est_files = sorted(glob.glob(EST_GLOB))
sco_files = sorted(glob.glob(SCORES_GLOB))

assert len(ref_files) == len(est_files) == len(sco_files) > 0

file_sets = list(zip(ref_files, est_files, sco_files))


@pytest.fixture
def transcription_data(request):
    ref_f, est_f, sco_f = request.param
    with open(sco_f) as f:
        expected_scores = json.load(f)
    # Load in an example segmentation annotation
    ref_int, ref_pitch = mir_eval.io.load_valued_intervals(ref_f)
    # Load in estimated transcription
    est_int, est_pitch = mir_eval.io.load_valued_intervals(est_f)

    return ref_int, ref_pitch, est_int, est_pitch, expected_scores


REF = np.array(
    [
        [0.100, 0.300, 220.000],
        [0.300, 0.400, 246.942],
        [0.500, 0.600, 277.183],
        [0.550, 0.650, 293.665],
    ]
)

EST = np.array(
    [
        [0.120, 0.290, 225.000],
        [0.300, 0.340, 246.942],
        [0.500, 0.600, 500.000],
        [0.550, 0.600, 293.665],
        [0.560, 0.650, 293.665],
    ]
)

SCORES = {
    "Precision": 0.4,
    "Recall": 0.5,
    "F-measure": 0.4444444444444445,
    "Average_Overlap_Ratio": 0.675,
    "Precision_no_offset": 0.6,
    "Recall_no_offset": 0.75,
    "F-measure_no_offset": 0.6666666666666665,
    "Average_Overlap_Ratio_no_offset": 0.5833333333333333,
}

ONSET_SCORES = {
    "Onset_Precision": 0.8,
    "Onset_Recall": 1.0,
    "Onset_F-measure": 0.8888888888888889,
}

OFFSET_SCORES = {
    "Offset_Precision": 0.6,
    "Offset_Recall": 0.75,
    "Offset_F-measure": 0.6666666666666665,
}


def test_match_note_offsets():
    ref_int = REF[:, :2]
    est_int = EST[:, :2]

    matching = mir_eval.transcription.match_note_offsets(ref_int, est_int)

    assert matching == [(0, 0), (2, 2), (3, 3)]


def test_match_note_offsets_strict():
    ref_int = REF[:, :2]
    est_int = EST[:, :2]

    matching = mir_eval.transcription.match_note_offsets(ref_int, est_int, strict=True)

    assert matching == [(0, 0), (2, 2), (3, 4)]


def test_match_note_onsets():
    ref_int = REF[:, :2]
    est_int = EST[:, :2]

    matching = mir_eval.transcription.match_note_onsets(ref_int, est_int)

    assert matching == [(0, 0), (1, 1), (2, 2), (3, 3)]


def test_match_note_onsets_strict():
    ref_int = REF[:, :2]
    est_int = EST[:, :2]

    matching = mir_eval.transcription.match_note_onsets(ref_int, est_int, strict=True)

    assert matching == [(0, 0), (1, 1), (2, 2), (3, 3)]


def test_match_notes():
    ref_int, ref_pitch = REF[:, :2], REF[:, 2]
    est_int, est_pitch = EST[:, :2], EST[:, 2]

    matching = mir_eval.transcription.match_notes(
        ref_int, ref_pitch, est_int, est_pitch
    )

    assert matching == [(0, 0), (3, 3)]

    matching = mir_eval.transcription.match_notes(
        ref_int, ref_pitch, est_int, est_pitch, offset_ratio=None
    )

    assert matching == [(0, 0), (1, 1), (3, 3)]


def test_match_notes_strict():
    ref_int, ref_pitch = np.array([[0, 1]]), np.array([100])
    est_int, est_pitch = np.array([[0.05, 1]]), np.array([100])

    matching = mir_eval.transcription.match_notes(
        ref_int, ref_pitch, est_int, est_pitch, strict=True
    )

    assert matching == []


def test_precision_recall_f1_overlap():
    # load test data
    ref_int, ref_pitch = REF[:, :2], REF[:, 2]
    est_int, est_pitch = EST[:, :2], EST[:, 2]

    (
        precision,
        recall,
        f_measure,
        avg_overlap_ratio,
    ) = mir_eval.transcription.precision_recall_f1_overlap(
        ref_int, ref_pitch, est_int, est_pitch
    )

    scores_gen = np.array([precision, recall, f_measure, avg_overlap_ratio])
    scores_exp = np.array(
        [
            SCORES["Precision"],
            SCORES["Recall"],
            SCORES["F-measure"],
            SCORES["Average_Overlap_Ratio"],
        ]
    )
    assert np.allclose(scores_exp, scores_gen, atol=A_TOL)

    (
        precision,
        recall,
        f_measure,
        avg_overlap_ratio,
    ) = mir_eval.transcription.precision_recall_f1_overlap(
        ref_int, ref_pitch, est_int, est_pitch, offset_ratio=None
    )

    scores_gen = np.array([precision, recall, f_measure, avg_overlap_ratio])
    scores_exp = np.array(
        [
            SCORES["Precision_no_offset"],
            SCORES["Recall_no_offset"],
            SCORES["F-measure_no_offset"],
            SCORES["Average_Overlap_Ratio_no_offset"],
        ]
    )
    assert np.allclose(scores_exp, scores_gen, atol=A_TOL)


def test_onset_precision_recall_f1():
    # load test data
    ref_int = REF[:, :2]
    est_int = EST[:, :2]

    precision, recall, f_measure = mir_eval.transcription.onset_precision_recall_f1(
        ref_int, est_int
    )

    scores_gen = np.array([precision, recall, f_measure])
    scores_exp = np.array(
        [
            ONSET_SCORES["Onset_Precision"],
            ONSET_SCORES["Onset_Recall"],
            ONSET_SCORES["Onset_F-measure"],
        ]
    )
    assert np.allclose(scores_exp, scores_gen, atol=A_TOL)


def test_offset_precision_recall_f1():
    # load test data
    ref_int = REF[:, :2]
    est_int = EST[:, :2]

    precision, recall, f_measure = mir_eval.transcription.offset_precision_recall_f1(
        ref_int, est_int
    )

    scores_gen = np.array([precision, recall, f_measure])
    scores_exp = np.array(
        [
            OFFSET_SCORES["Offset_Precision"],
            OFFSET_SCORES["Offset_Recall"],
            OFFSET_SCORES["Offset_F-measure"],
        ]
    )
    assert np.allclose(scores_exp, scores_gen, atol=A_TOL)


@pytest.mark.parametrize("transcription_data", file_sets, indirect=True)
def test_regression(transcription_data):
    ref_int, ref_pitch, est_int, est_pitch, expected_scores = transcription_data

    scores = mir_eval.transcription.evaluate(ref_int, ref_pitch, est_int, est_pitch)
    assert scores.keys() == expected_scores.keys()
    for metric in scores:
        assert np.allclose(scores[metric], expected_scores[metric], atol=A_TOL)


@pytest.mark.xfail(raises=ValueError)
@pytest.mark.parametrize(
    "ref_pitch, est_pitch",
    [(np.array([-100]), np.array([100])), (np.array([100]), np.array([-100]))],
)
def test_invalid_pitch(ref_pitch, est_pitch):
    ref_int = np.array([[0, 1]])
    mir_eval.transcription.validate(ref_int, ref_pitch, ref_int, est_pitch)


@pytest.mark.xfail(raises=ValueError)
@pytest.mark.parametrize(
    "ref_int, est_int",
    [
        (np.array([[0, 1], [2, 3]]), np.array([[0, 1]])),
        (np.array([[0, 1]]), np.array([[0, 1], [2, 3]])),
    ],
)
def test_inconsistent_int_pitch(ref_int, est_int):
    ref_pitch = np.array([100])
    mir_eval.transcription.validate(ref_int, ref_pitch, est_int, ref_pitch)


def test_empty_ref():
    ref_int, ref_pitch = np.empty(shape=(0, 2)), np.array([])
    est_int, est_pitch = np.array([[0, 1]]), np.array([100])

    with pytest.warns(UserWarning, match="Reference notes are empty"):
        mir_eval.transcription.validate(ref_int, ref_pitch, est_int, est_pitch)


def test_empty_est():
    ref_int, ref_pitch = np.array([[0, 1]]), np.array([100])
    est_int, est_pitch = np.empty(shape=(0, 2)), np.array([])

    with pytest.warns(UserWarning, match="Estimated notes are empty"):
        mir_eval.transcription.validate(ref_int, ref_pitch, est_int, est_pitch)


@pytest.mark.filterwarnings("ignore:.*notes are empty")
def test_precision_recall_f1_overlap_empty():
    ref_int, ref_pitch = np.empty(shape=(0, 2)), np.array([])
    est_int, est_pitch = np.array([[0, 1]]), np.array([100])

    (
        precision,
        recall,
        f1,
        avg_overlap_ratio,
    ) = mir_eval.transcription.precision_recall_f1_overlap(
        ref_int, ref_pitch, est_int, est_pitch
    )

    assert (precision, recall, f1) == (0, 0, 0)

    (
        precision,
        recall,
        f1,
        avg_overlap_ratio,
    ) = mir_eval.transcription.precision_recall_f1_overlap(
        est_int, est_pitch, ref_int, ref_pitch
    )

    assert (precision, recall, f1) == (0, 0, 0)


@pytest.mark.filterwarnings("ignore:.*notes are empty")
def test_onset_precision_recall_f1_empty():
    ref_int = np.empty(shape=(0, 2))
    est_int = np.array([[0, 1]])

    precision, recall, f1 = mir_eval.transcription.onset_precision_recall_f1(
        ref_int, est_int
    )

    assert (precision, recall, f1) == (0, 0, 0)

    precision, recall, f1 = mir_eval.transcription.onset_precision_recall_f1(
        est_int, ref_int
    )

    assert (precision, recall, f1) == (0, 0, 0)


@pytest.mark.filterwarnings("ignore:.*notes are empty")
def test_offset_precision_recall_f1_empty():
    ref_int = np.empty(shape=(0, 2))
    est_int = np.array([[0, 1]])

    precision, recall, f1 = mir_eval.transcription.offset_precision_recall_f1(
        ref_int, est_int
    )

    assert (precision, recall, f1) == (0, 0, 0)

    precision, recall, f1 = mir_eval.transcription.offset_precision_recall_f1(
        est_int, ref_int
    )

    assert (precision, recall, f1) == (0, 0, 0)



"""
Unit tests for mir_eval.alignment
"""

import glob
import json

import pytest
import numpy as np

import mir_eval

A_TOL = 1e-12

# Path to the fixture files
REF_GLOB = "data/alignment/ref*.txt"
EST_GLOB = "data/alignment/est*.txt"
SCORES_GLOB = "data/alignment/output*.json"

ref_files = sorted(glob.glob(REF_GLOB))
est_files = sorted(glob.glob(EST_GLOB))
sco_files = sorted(glob.glob(SCORES_GLOB))

assert len(ref_files) == len(est_files) == len(sco_files) > 0

file_sets = list(zip(ref_files, est_files, sco_files))


@pytest.mark.xfail(raises=ValueError)
@pytest.mark.parametrize(
    "metric",
    [
        mir_eval.alignment.absolute_error,
        mir_eval.alignment.percentage_correct,
        mir_eval.alignment.percentage_correct_segments,
        (
            lambda ref_ts, est_ts: mir_eval.alignment.percentage_correct_segments(
                ref_ts, est_ts, duration=max(np.max(ref_ts), np.max(est_ts))
            )
        ),
        mir_eval.alignment.karaoke_perceptual_metric,
    ],
)
@pytest.mark.parametrize(
    "est_alignment, pred_alignment",
    [
        (
            np.array([[1.0, 2.0]]),
            np.array([[1.0, 2.0]]),
        ),  # alignments must be 1d ndarray
        (
            np.array([[-1.0, 2.0]]),
            np.array([[1.0, 2.0]]),
        ),  # alignments must be non-negative
        (np.array([[2.0, 1.0]]), np.array([[1.0, 2.0]])),  # alignments must be sorted
        (
            np.array([[1.0, 2.0]]),
            np.array([[1.0]]),
        ),  # alignments must have the same length
    ],
)
def test_alignment_functions_fail(metric, est_alignment, pred_alignment):
    metric(est_alignment, pred_alignment)


@pytest.fixture
def alignment_data(request):
    ref_f, est_f, sco_f = request.param
    with open(sco_f) as f:
        expected_scores = json.load(f)
    reference_alignments = mir_eval.io.load_events(ref_f)
    estimated_alignments = mir_eval.io.load_events(est_f)

    return reference_alignments, estimated_alignments, expected_scores


@pytest.mark.parametrize("alignment_data", file_sets, indirect=True)
def test_alignment_functions(alignment_data):
    reference_alignments, estimated_alignments, expected_scores = alignment_data
    scores = mir_eval.alignment.evaluate(reference_alignments, estimated_alignments)

    assert scores.keys() == expected_scores.keys()
    for metric in scores:
        assert np.allclose(scores[metric], expected_scores[metric], atol=A_TOL)



#!/usr/bin/env python
"""
Unit tests for mir_eval.tempo
"""
import numpy as np
import mir_eval
import json
import glob
import pytest


A_TOL = 1e-12


REF_GLOB = "data/tempo/ref*.lab"
EST_GLOB = "data/tempo/est*.lab"
SCORES_GLOB = "data/tempo/output*.json"

ref_files = sorted(glob.glob(REF_GLOB))
est_files = sorted(glob.glob(EST_GLOB))
sco_files = sorted(glob.glob(SCORES_GLOB))

assert len(ref_files) == len(est_files) == len(sco_files) > 0

file_sets = list(zip(ref_files, est_files, sco_files))


@pytest.fixture
def tempo_data(request):
    ref_f, est_f, sco_f = request.param
    with open(sco_f) as f:
        expected_scores = json.load(f)

    def _load_tempi(filename):
        values = mir_eval.io.load_delimited(filename, [float] * 3)
        return np.concatenate(values[:2]), values[-1][0]

    reference_tempi, ref_weight = _load_tempi(ref_f)
    estimated_tempi, _ = _load_tempi(est_f)

    return reference_tempi, ref_weight, estimated_tempi, expected_scores


def test_zero_tolerance_pass():
    good_ref = np.array([60, 120])
    good_weight = 0.5
    good_est = np.array([120, 180])
    zero_tol = 0.0

    with pytest.warns(
        UserWarning, match="A tolerance of 0.0 may not lead to the results you expect"
    ):
        mir_eval.tempo.detection(good_ref, good_weight, good_est, tol=zero_tol)


def test_tempo_pass():
    good_ref = np.array([60, 120])
    good_weight = 0.5
    good_est = np.array([120, 180])
    good_tol = 0.08

    for good_tempo in [np.array([50, 50]), np.array([0, 50]), np.array([50, 0])]:
        mir_eval.tempo.detection(good_tempo, good_weight, good_est, good_tol)
        mir_eval.tempo.detection(good_ref, good_weight, good_tempo, good_tol)

    # allow both estimates to be zero
    mir_eval.tempo.detection(good_ref, good_weight, np.array([0, 0]), good_tol)


@pytest.mark.xfail(raises=ValueError)
def test_tempo_zero_ref():
    # Both references cannot be zero
    mir_eval.tempo.detection(np.array([0.0, 0.0]), 0.5, np.array([60, 120]))


@pytest.mark.xfail(raises=ValueError)
@pytest.mark.parametrize("weight", [-1, 1.5])
def test_tempo_weight_range(weight):
    # Weight needs to be in the range [0, 1]
    mir_eval.tempo.detection(np.array([60, 120]), weight, np.array([120, 180]))


@pytest.mark.xfail(raises=ValueError)
@pytest.mark.parametrize("tol", [-1, 1.5])
def test_tempo_tol_range(tol):
    # Weight needs to be in the range [0, 1]
    mir_eval.tempo.detection(np.array([60, 120]), 0.5, np.array([120, 180]), tol=tol)


@pytest.mark.xfail(raises=ValueError)
@pytest.mark.parametrize(
    "bad_tempo",
    [
        np.array([-1, -1]),
        np.array([-1, 0]),
        np.array([-1, 50]),
        np.array([0, 1, 2]),
        np.array([0]),
    ],
)
def test_tempo_fail_bad_reftempo(bad_tempo):
    good_ref = np.array([60, 120])
    good_est = np.array([120, 180])

    mir_eval.tempo.detection(bad_tempo, 0.5, good_est, 0.08)


@pytest.mark.xfail(raises=ValueError)
@pytest.mark.parametrize(
    "bad_tempo",
    [
        np.array([-1, -1]),
        np.array([-1, 0]),
        np.array([-1, 50]),
        np.array([0, 1, 2]),
        np.array([0]),
    ],
)
def test_tempo_fail_bad_esttempo(bad_tempo):
    good_ref = np.array([60, 120])
    good_est = np.array([120, 180])

    mir_eval.tempo.detection(good_ref, 0.5, bad_tempo, 0.08)


@pytest.mark.parametrize("tempo_data", file_sets, indirect=True)
def test_tempo_regression(tempo_data):
    ref_tempi, ref_weight, est_tempi, expected_scores = tempo_data

    scores = mir_eval.tempo.evaluate(ref_tempi, ref_weight, est_tempi)
    assert scores.keys() == expected_scores.keys()
    for metric in scores:
        assert np.allclose(scores[metric], expected_scores[metric], atol=A_TOL)
